/*	$Id: codawrap.c,v 1.61 2006-08-21 16:27:23 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/


//
// codawrap
// OSR call interface

#include <coda_types.h>
#include <coda.h>
#include <coda_nt.h>
#include <cnode.h>

// Access to root file handle.

static coda_filehandle *root_fh;
static int root_init = FALSE;
static char DosDrive = 'C';

#ifdef _BIG_FILE_
static const LONG megabyte = 1048576;
#endif

// Static data -- the standard .lnk header that works ...

static lnk_head lnk_header =
    {0x4C, 0x1, 0x14, 0x2, 0, 0, 0, 0, 0, 0xc0, 0, 0, 0, 0, 0, 0, 0x46,
     0x0c, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0 };


// prototypes needed.

NTSTATUS coda_internal_lookup(IN coda_inode *dirinode,
			      IN PCHAR FileName,
			      IN int NameLen,
			      IN BOOLEAN CaseInsensitive,
			      IN int Uid,
			      IN BOOLEAN CheckForBF,
			      IN BOOLEAN FindSymLinks,
			      OUT coda_inode **fileinode,
			      OUT coda_filehandle **pfh);

NTSTATUS coda_internal_getattr(IN coda_inode *inode,
			       IN int Uid,
			       IN BOOLEAN CheckForBF,
			       IN OUT PFS_FILE_ATTRIBUTES FileAttributes);

//
// Utility routine to purge all file handles on an inode
//

static void
purge_fh_list (IN coda_inode *inode)
{
    coda_filehandle *tfh, *nfh;

    // Get the head
    nfh = inode->c_fh_list;

    // Move down the list ...
    while (nfh != NULL) {
	tfh = nfh;			// Done in this order to make sure
	nfh = nfh->fh_next;		// tfh can be removed from list.
	CDEBUG((D_WRAPPER, "purge_fh_list: flush wrapper cache fh %d (%s)\n",
		tfh->fh_unique, inode->c_name));
	if (!NT_SUCCESS(OwPurgeCache(tfh)))
	    DbgPrint ("purge_fh_list: can't purge file handle cache for %s\n",
		      inode->c_name);
    }
}


// Wrapper calls

NTSTATUS coda_mount(PVOID MediaHandle, PFS_VOL_HANDLE CodaVolumep,
		    PFS_FILE_HANDLE CodaRootDirHandle)
{
    int CodaInfo = (int)MediaHandle;
    coda_inode *codaroot;
    int len; 
    int rc;
    CodaFid fid;

    coda_inode      *ctl_inode;
    coda_filehandle *ctl_fh;
    CodaFid          ctl_fid = CTL_FID;

    NTSTATUS status;
    int Uid = GetNtUID();

#ifdef DBG
    clean_msg_list();
#endif
    
    ENTRY(D_WRAPPER, "coda_mount");
    
    if ( CODA_MOUNTED(&coda_nt_volume) ) 
	return STATUS_UNSUCCESSFUL;


    if ( (CodaInfo & 0xffffff00) == (('C'<<24)+('F'<<16)+('S'<<8))) {
	DosDrive = (CodaInfo & 0xff);
    } else {
	CDEBUG((D_WRAPPER, "coda_mount: Old info!\n"));
	//	return STATUS_UNSUCCESSFUL;
    }


    // Save the coda device object

    coda_device_object = (PDEVICE_OBJECT) *CodaVolumep;

  
    // Build the root inode and file handle ...

    rc = venus_rootfid(&fid);
    if ( rc ) {
	CDEBUG((D_WRAPPER, "coda_mount: Venus error: %d\n", rc));
	return STATUS_UNRECOGNIZED_VOLUME;
    }
    
    coda_nt_volume.v_refcount = 0;

    codaroot = inode_new(&fid, NULL, "\\");
    if (!codaroot) 
	return STATUS_INSUFFICIENT_RESOURCES;

    inode_pin (codaroot);
    root_fh = get_cfh (codaroot, codaroot, Uid);
    if (!root_fh)
	return STATUS_INSUFFICIENT_RESOURCES;
	
    CDEBUG((D_WRAPPER, "coda_mount: allocated coda_inode! fh %d\n",
	    root_fh->fh_unique));
    
    rc = coda_internal_getattr(codaroot, root_fh->fh_uid, FALSE,
			       &codaroot->c_fattr);

    if ( rc ) {
	CDEBUG((D_WRAPPER, "coda_mount: error %d getting attributes.\n", rc));
	return STATUS_UNRECOGNIZED_VOLUME;
    }

    ctl_inode = inode_new (&ctl_fid, codaroot, ".CONTROL");
    if (!ctl_inode)
	return STATUS_INSUFFICIENT_RESOURCES;

    inode_pin (ctl_inode);
    ctl_fh = get_cfh (ctl_inode, codaroot, Uid);
    if (!ctl_fh)
	return STATUS_INSUFFICIENT_RESOURCES;

    // Set the "attributes" on the .CONTROL file.
    
    RtlSecondsSince1970ToTime(100000000, &ctl_inode->c_fattr.CreationTime);
    RtlSecondsSince1970ToTime(100000000, &ctl_inode->c_fattr.LastAccessTime);
    RtlSecondsSince1970ToTime(100000000, &ctl_inode->c_fattr.LastModifiedTime);
    ctl_inode->c_fattr.ValidDataLength.QuadPart = 0;
    ctl_inode->c_fattr.DiskLength.QuadPart = 0;
    ctl_inode->c_fattr.Attributes = FILE_ATTRIBUTE_NORMAL;
    ctl_inode->c_fstype = OSRFS_TYPE_FILE;
    inode_setflag (ctl_inode, I_VATTR);

    // Set the coda volume information

    coda_nt_volume.v_namelen = 0;
    coda_nt_volume.v_name = NULL;
    coda_nt_volume.v_rootinode = codaroot;
    coda_nt_volume.v_rootfh = root_fh;
    coda_nt_volume.v_ctlinode = ctl_inode;
    coda_nt_volume.v_ctlfh = ctl_fh;
    coda_nt_volume.v_status = V_MOUNTED | V_ROOT;
    
    *CodaRootDirHandle = (FS_FILE_HANDLE) root_fh;
    *CodaVolumep = (FS_VOL_HANDLE) &coda_nt_volume;

    CDEBUG((D_WRAPPER, "coda_mount: Volume is 0x%x, DosDrive = '%c'\n",
	    &coda_nt_volume, DosDrive));
    
    EXIT(D_WRAPPER, "coda_mount");
    return STATUS_SUCCESS;
}

NTSTATUS coda_unmount(IN FS_VOL_HANDLE CodaVolume)
{
    struct coda_ntvolume *Volume = (struct coda_ntvolume *)CodaVolume;

    DbgPrint ("coda_umount: entry!\n");

    // ENTRY(D_WRAPPER, "coda_unmount");

    if (Volume != &coda_nt_volume) {
      CDEBUG((D_WRAPPER, "coda_unmount: Wrong Volume!\n"));
      // return some error code
      return STATUS_INVALID_PARAMETER_1;
    }
    
    // Clear request queue

    
#ifdef DBG
    clean_msg_list();
#endif

    // Free all objects allocated .. ???  XXX
    //    coda_release (coda_nt_volume.v_ctlfh);
    //    coda_release (coda_nt_volume.v_rootfh);

    inode_freeall();
    
    coda_nt_volume.v_rootinode = NULL;
    coda_nt_volume.v_rootfh  = NULL;
    coda_nt_volume.v_ctlinode = NULL;
    coda_nt_volume.v_ctlfh  = NULL;
    coda_nt_volume.v_status = 0;

    if (coda_nt_volume.v_refcount != 1) {
      CDEBUG((D_WRAPPER, "coda_unmount: refs=%d!\n",
	      coda_nt_volume.v_refcount));
    }

    EXIT(D_WRAPPER, "coda_unmount");
    return STATUS_SUCCESS;
}


//
//  coda_verify -- is this a Coda Volume ?
//

NTSTATUS coda_verify(IN FS_VOL_HANDLE CodaVolume)
{
    struct coda_ntvolume *Volume = (struct coda_ntvolume *)CodaVolume;

    ENTRY(D_SUPER, "coda_verify");
    
    if (Volume != &coda_nt_volume || !CODA_ROOTVOL(Volume)
	|| !CODA_MOUNTED(Volume)) {
	CDEBUG((D_WRAPPER, "coda_verify: returning wrong volume.\n"));
	CDEBUG((D_WRAPPER, "    :Volume = 0x%x, coda_nt_volume = 0x%x\n",
	       Volume, &coda_nt_volume));
	return STATUS_WRONG_VOLUME;
    }

    EXIT(D_SUPER, "coda_verify");
    return STATUS_SUCCESS;
}


//
// coda_get_volume_attributes
//

NTSTATUS coda_get_volume_attributes(IN FS_VOL_HANDLE CodaVolume,
				    OUT PFS_VOL_ATTRIBUTES Attributes,
				    PULONG BufferSize)
{
    struct coda_ntvolume *Volume = (struct coda_ntvolume *)CodaVolume;
    NTSTATUS Status;
    ANSI_STRING AnsiString;
    UNICODE_STRING UnicodeString;
    struct coda_statfs statfs;
    int rc;

    WCHAR vname[10] = L"Coda";
    
    ENTRY(D_LOOKATTR, "coda_get_volume_attributes");

    if (Volume != &coda_nt_volume) {
	CDEBUG((D_WRAPPER, "coda_get_volume_attributes: Wrong Volume!\n"));
	CDEBUG((D_WRAPPER, "    :Volume = 0x%x, coda_nt_volume = 0x%x\n",
	       Volume, &coda_nt_volume));
	// Return an error?
    }
  
    //
    // If this is a call to get the attributes for the UNC root
    // volume... we just make a something.
    //
    if( ! CODA_ROOTVOL(Volume) ) {
	DbgPrint ("coda_get_volume_attributes: Non-Root volume!\n");
    }

    //
    // Clear the attributes are we are returning.
    //
    RtlZeroMemory(Attributes, sizeof(FS_VOL_ATTRIBUTES) + 10 * sizeof(WCHAR));

    // 
    // Get the cache statistics from venus
    //

    rc = venus_statfs (&statfs);
    if (rc) {
	Attributes->VolumeSize.HighPart = 0;
	Attributes->VolumeSize.LowPart = 72727272;
	Attributes->FreeSpace.HighPart = 0;
	Attributes->FreeSpace.LowPart = 72727272;
    } else {
     //	Attributes->VolumeSize.QuadPart = statfs.f_blocks; 
     //	Attributes->FreeSpace.QuadPart = statfs.f_bavail;
     Attributes->VolumeSize.QuadPart = 1000000000;
     Attributes->FreeSpace.QuadPart =   500000000 -
       (statfs.f_blocks - statfs.f_bavail);
    }
    
    Attributes->CreationTime = Volume->v_rootinode->c_fattr.CreationTime;
    Attributes->BytesPerAllocationUnit = 1024;
    Attributes->DeviceType = FILE_DEVICE_NETWORK_FILE_SYSTEM; // Was _DISK
    Attributes->DeviceCharacteristics =
	FILE_REMOTE_DEVICE | FILE_CASE_SENSITIVE_SEARCH
	| FILE_CASE_PRESERVED_NAMES;

    memcpy(&Attributes->VolumeLabel[0], vname, wcslen(vname) * sizeof(WCHAR));
    Attributes->LabelLength = wcslen(vname) * sizeof(WCHAR);
    
    EXIT(D_WRAPPER, "coda_get_volume_attributes");
    return(STATUS_SUCCESS);
}


//
// coda_release   XXX needs a lot more work!
//

NTSTATUS coda_release(FS_FILE_HANDLE handle)
{
    coda_filehandle *fh = (coda_filehandle *) handle;
    coda_inode *inode = fh->fh_inode;
    struct coda_ntvolume *ntvol;
    NTSTATUS rc =0;
    int flags;
    
    ENTRY(D_LOOKATTR, "coda_release");
    rc = STATUS_SUCCESS;
    
    if (!CODA_ISOURS(inode)) {
	return STATUS_INVALID_PARAMETER_1;
    }

    ntvol = inode->c_ntvol;
    if (!CODA_NTVOL(ntvol)) {
	return STATUS_UNRECOGNIZED_VOLUME;    
    }
    
    CDEBUG((D_LOOKATTR, "coda_release: for %s %s (fh %d%s)\n", inode->c_name,
	    coda_f2s(&inode->c_fid), fh->fh_unique,
	    fh->fh_isopen ? " is open": ""));

    if (fh == coda_nt_volume.v_ctlfh) {
	CDEBUG ((D_LOOKATTR, "coda_release: control fh, returning success\n"));
	return STATUS_SUCCESS;
    }

    LockInode (inode);

    CDEBUG((D_LOOKATTR, "coda_release: freeing fh %d (type %d).\n",
	    fh->fh_unique, inode->c_fstype));

   
#ifdef _BIG_FILE_
    // If a big file, the fh of the top level file won't be open, so
    // just release the hunk files that may be open below. 
    if (inode->c_big_filep) {
	int temp;
	CDEBUG((D_LOOKATTR, "coda_release: closing big file.\n"));
	// Release the current open file and dirs
	for (temp = 0; temp < 3; temp++) {
	    if (inode->c_big_filep->bf_curr.hu_fh[temp]) {
		inode_setflag(inode->c_big_filep->bf_curr.hu_fh[temp]->fh_inode,
			      I_PURGE);
		coda_release (inode->c_big_filep->bf_curr.hu_fh[temp]);
		if (inode->c_big_filep->bf_curr.hu_fh[temp] == 
		    inode->c_big_filep->bf_next.hu_fh[temp])
		    inode->c_big_filep->bf_next.hu_fh[temp] = NULL;
		inode->c_big_filep->bf_curr.hu_fh[temp] = NULL;
	    }
	    if (inode->c_big_filep->bf_next.hu_fh[temp]) {
		inode_setflag(inode->c_big_filep->bf_next.hu_fh[temp]->fh_inode,
			      I_PURGE);
		coda_release (inode->c_big_filep->bf_next.hu_fh[temp]);
		inode->c_big_filep->bf_next.hu_fh[temp] = NULL;
	    }
	}
	inode->c_big_filep->bf_curr.hu_num = -1;
	inode->c_big_filep->bf_next.hu_num = -1;
    }
#endif

    if (fh->fh_isopen) {
	if (!(--(inode->c_open_cnt))) {
	    CDEBUG((D_FILE, "coda_release: Closing inode file object.\n")); 
	    

	    // Close the object!
	    if (inode->c_object != NULL) {
		inode_close_handle(inode);
	    } else {
		DbgPrint("coda_release: Closing NULL c_object\n");
	    }
	}
	
	// Tell Venus
	rc = venus_close(&inode->c_fid, fh->fh_openflags, fh->fh_uid);
	
	// ignore error
	if ( rc ) {
	    DbgPrint("coda_release: venus_close error %d\n", rc);
	}
	
	// being paranoid ...
	fh->fh_isopen = 0;
    }

    //  Change .lnk files to read-only, these are regular files, not links
    //  files to venus.
    if (strcasetail (inode->c_name, ".lnk")) {
	inode->c_fattr.Attributes = FILE_ATTRIBUTE_READONLY;
	purge_fh_list (inode);
    }
    
    if (!inode_hasflag (inode, I_PINNED)) {
	CDEBUG((D_WRAPPER, "coda_release: releasing unpinned inode\n"));
    }

    UnLockInode(inode);
 
    free_cfh (fh);

    EXIT (D_LOOKATTR, "coda_release");
    return rc;
}

// 
// Local helper routine
//

static void coda_vattr2FATTR(struct coda_vattr *Attributes,
			     PFS_FILE_ATTRIBUTES FileAttributes, PULONG Type)
{
    RtlZeroMemory(FileAttributes, sizeof(FS_FILE_ATTRIBUTES));

    //
    // Do the conversions for the data being send back
    //
    RtlSecondsSince1970ToTime(Attributes->va_atime.ts_sec,
			      &FileAttributes->CreationTime);
    RtlSecondsSince1970ToTime(Attributes->va_atime.ts_sec,
			      &FileAttributes->LastAccessTime);
    RtlSecondsSince1970ToTime(Attributes->va_mtime.ts_sec,
			      &FileAttributes->LastModifiedTime);
    
    //
    // File size
    //
    FileAttributes->ValidDataLength.QuadPart = Attributes->va_size;

    //
    // Allocation size
    //
    FileAttributes->DiskLength.QuadPart = Attributes->va_size;
    
    // link count
    FileAttributes->LinkCount = Attributes->va_nlink;
    
    // 
    // File type
    //
    
    switch(Attributes->va_type) {
    case C_VDIR:
	*Type = OSRFS_TYPE_DIRECTORY;
	FileAttributes->Attributes = FILE_ATTRIBUTE_DIRECTORY;
	
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: File is a directory\n"));
	break;
	
    case C_VLNK:
	*Type = OSRFS_TYPE_SYMBOLIC_LINK;
	FileAttributes->Attributes = FILE_ATTRIBUTE_READONLY;
	
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: File is a link\n"));
	break;
	
    case C_VREG:
	*Type = OSRFS_TYPE_FILE;
	if ((Attributes->va_mode & C_M_WRITE))
	    FileAttributes->Attributes = FILE_ATTRIBUTE_NORMAL;
	else
	    FileAttributes->Attributes = FILE_ATTRIBUTE_READONLY;

	
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: File is a regular file\n"));
	break;
	
    default :
	*Type = OSRFS_TYPE_OTHER;
	CDEBUG((D_LOOKATTR, "coda_vattr2FATTR: unknown file type!\n"));
    }
    
    return;
    
}

//
// coda_internal_getattr using inodes. (was coda_getattr)
// 

NTSTATUS coda_internal_getattr(IN coda_inode *inode,
			       IN int Uid,
			       IN BOOLEAN CheckForBF,
			       IN OUT PFS_FILE_ATTRIBUTES FileAttributes)
{
    NTSTATUS rc;
    struct coda_ntvolume *ntvol;
    struct coda_vattr Attributes;

    CDEBUG ((D_LOOKATTR, "coda_internal_getattr: Starting on %s: %s, uid %d\n",
	     inode->c_name, coda_f2s(&inode->c_fid), Uid));

    LockInode(inode);
    
    try {

	//
	// Do some sanity checking to be sure that the handle that
	// was passed in is really a valid file handle.
	//
	if (! CODA_ISOURS(inode) )  {
	    CDEBUG((D_WRAPPER, "coda_internal_getattr: NOT A FILE!\n"));
	    return(STATUS_INVALID_PARAMETER_1);
	}
	
	//
	// Zero attributes buffer before anything else.  This pre-sets
	// all fields to zero.  The wrapper wants us to pass back any
	// fields we do not support as zero.
	//
	if (FileAttributes != NULL) {
	    __try {
		RtlZeroMemory(FileAttributes, sizeof(FS_FILE_ATTRIBUTES));
	    }
	    except(EXCEPTION_EXECUTE_HANDLER) {
		CDEBUG((D_WRAPPER, "coda_getattar: can't clear buffer!\n"));
		return(STATUS_INVALID_USER_BUFFER);
	    }
	}
	
	/* likely we are lucky and have this stuff already */
	if (!inode_hasflag(inode, I_VATTR)) {

	    rc = venus_getattr(&inode->c_fid, &Attributes, Uid);
	    
	    if (!NT_SUCCESS(rc)) {
		CDEBUG((D_LOOKATTR, "coda_internal_getattr: venus_getattr %s "
		       "returns error %d\n", coda_f2s(&inode->c_fid), rc));
		return(rc);
	    }
 
	    /* fill in return value and store result in coda_inode */
	    coda_vattr2FATTR(&Attributes, &inode->c_fattr, &inode->c_fstype);

	    // *sigh* ... make it look like a normal symlink
	    if (inode->c_fstype == OSRFS_TYPE_SYMBOLIC_LINK) {
		inode->c_fstype = OSRFS_TYPE_FILE;
		inode->c_lnk_state = SLNK_NORMAL;
	    }

	    // Get symbolic link name if needed.
	    if (CODA_ISSYMLINK(inode)) {
		char *tmpptr;
		int ix;
		int namelen = CODA_MAXNAMLEN-1;
		int codaprefix = 0;

		if (inode->c_lnk_state == SLNK_NEW) {
		    // Newly created link or empty link => empty file
		    CDEBUG((D_LOOKATTR, "coda_internal_getattr: "
			    "New symbolic link, make empty.\n"));
		    // set_length needs to change this value
		    inode->c_fattr.LastByteWritten.QuadPart
			= inode->c_fattr.ValidDataLength.QuadPart
			= inode->c_fattr.DiskLength.QuadPart
			= inode->c_lnk_len
			= 0;
		    if (inode->c_lnk_file != NULL) {
			coda_free(inode->c_lnk_file);
			inode->c_lnk_file = NULL;
		    }
		    // To allow writing of new links.
		    inode->c_fattr.Attributes = FILE_ATTRIBUTE_NORMAL;

		} else {

		    rc = venus_readlink (&inode->c_fid, inode->c_containername,
					 &namelen, Uid);
		    
		    if (!NT_SUCCESS(rc)) {
			CDEBUG((D_LOOKATTR,
				"coda_internal_getattr: venus_readlink %s "
				"returns error %d\n", coda_f2s(&inode->c_fid),
				rc));
			return(rc);
		    }
		    
		    CDEBUG((D_LOOKATTR, "coda_internal_getattr: symlink to %s"
			    " (namelen = %d)\n", inode->c_containername,
			    namelen));
		    
		    if (StrNCmp(inode->c_containername, "/coda/", 4) == 0)
			codaprefix = 1;
		    
		    // Gin up the fake file 
		    inode->c_fattr.LastByteWritten.QuadPart
			= inode->c_fattr.ValidDataLength.QuadPart
			= inode->c_fattr.DiskLength.QuadPart
			= inode->c_lnk_len
			= sizeof(lnk_head) + 4 + 2*namelen - 3*codaprefix;
		    inode->c_fattr.Attributes = FILE_ATTRIBUTE_READONLY ;
		    if (inode->c_lnk_file != NULL)
			coda_free(inode->c_lnk_file);
		    inode->c_lnk_file
			= (char *)coda_malloc(inode->c_lnk_len+1);
		    if (inode->c_lnk_file) {
			CDEBUG((D_LOOKATTR,
				"coda_internal_getattr: Building .lnk file: "
				" lnk_len = %d\n", inode->c_lnk_len));
			inode->c_lnk_state = SLNK_NORMAL;
			memcpy(inode->c_lnk_file, &lnk_header,
			       sizeof(lnk_header));
			tmpptr = inode->c_lnk_file + sizeof(lnk_header);
			*((short *)tmpptr) = (short)namelen;
			tmpptr += 2;
			memcpy(tmpptr, inode->c_containername, namelen);
			tmpptr += namelen;
			if (codaprefix) {
			    namelen -= 3;
			    *((short *)tmpptr) = (short)namelen;
			    tmpptr += 2;
			    tmpptr[0] = DosDrive;
			    tmpptr[1] = ':';
			    memcpy(tmpptr+2, inode->c_containername+5,
				   namelen-2);
			    tmpptr[namelen] = 0;
			    CDEBUG((D_LOOKATTR, "coda_internal_getattr: "
				    "Dos path '%s'\n", tmpptr));
			} else {
			    *((short *)tmpptr) = (short)namelen;
			    tmpptr += 2;
			    memcpy(tmpptr, inode->c_containername, namelen);
			}
			for (ix = 0; ix < namelen; ix++)
			    if (tmpptr[ix] == '/')
				tmpptr[ix] = '\\';
		    } else {
			CDEBUG((D_LOOKATTR, "coda_internal_getattr: "
				"link file malloc error\n"));
			inode->c_lnk_len = 0;
		    }
		}
	    }
	    
	    inode_setflag (inode, I_VATTR);

#ifdef _BIG_FILE_
	    // Big File Support -- look for _Coda_BigFile_ in a directory 
	    if (CheckForBF && CODA_ISDIR (inode) &&
		(inode != root_fh->fh_inode)) {

		// Look for the magic entry
		coda_inode *bfinfo;
		coda_filehandle *bfifh;
		NTSTATUS status
		    = coda_internal_lookup (inode, "_Coda_BigFile_", 14,
					    FALSE, Uid, FALSE, FALSE,
					  &bfinfo, &bfifh);

		if (NT_SUCCESS(status) && !CODA_ISDIR(bfinfo)) {
		    char *TempData;
		    PMDL pmdl;
		    ULONG BytesRead;

		    CDEBUG ((D_LOOKATTR,
			     "coda_internal_getattr: big file found.\n"));

		    // Read data

		    TempData = coda_malloc(128);
		    if (!TempData) {
			CDEBUG ((D_LOOKATTR,
				 "coda_internal_getattr: can't allocate "
				 "big file Buffer\n"));
			coda_release (bfifh);
			return STATUS_INSUFFICIENT_RESOURCES;      
		    }

		    pmdl=IoAllocateMdl (TempData, 128, FALSE, FALSE, NULL);
		    if (!pmdl) { 
			CDEBUG ((D_LOOKATTR,
				 "coda_internal_getattr: can't allocate "
				 "big file MDL\n"));
			coda_release (bfifh);
			return STATUS_INSUFFICIENT_RESOURCES;      
		    }
		    MmBuildMdlForNonPagedPool (pmdl);

		    status = coda_read(bfifh, RtlConvertUlongToLargeInteger(0),
				       pmdl, 127, &BytesRead);
		    
		    if (NT_SUCCESS(status)) {
			char *tmp;
			TempData[BytesRead] = 0;
			CDEBUG((D_LOOKATTR, "bigfile data: %s", TempData));
			if (StrNCmp(TempData, "CDBGFL00", 8) == 0) {
			    int temp;
			    coda_big_file *bfp;
			    // This is a big file!
			    inode->c_big_filep
				= coda_malloc(sizeof(coda_big_file));
			    bfp = inode->c_big_filep;
			    tmp = &TempData[8];
			    bfp->bf_size.QuadPart = A2Quad(tmp);
			    while (*tmp == ' ') tmp++;
			    while (IsDigit(*tmp)) tmp++;
			    bfp->bf_hunkbytes = megabyte * A2Long(tmp);
			    while (*tmp == ' ') tmp++;
			    while (IsDigit(*tmp)) tmp++;
			    bfp->bf_filesperdir = A2Long(tmp);
			    while (*tmp == ' ') tmp++;
			    while (IsDigit(*tmp)) tmp++;
			    bfp->bf_files = A2Long(tmp);

			    bfp->bf_curr.hu_num = -1;
			    bfp->bf_next.hu_num = -1;
			    for (temp = 0; temp < 3; temp++) {
				bfp->bf_curr.hu_fh[temp] = NULL;
				bfp->bf_next.hu_fh[temp] = NULL;
			    }

			    bfp->bf_dirdigits = 1;
			    for (temp = (bfp->bf_filesperdir - 1)/10; temp > 0;
				 temp /= 10)
				bfp->bf_dirdigits ++;
			    
			    temp = (bfp->bf_files-1)/bfp->bf_filesperdir;
			    bfp->bf_levels = 0;
			    while (temp > 0) {
				bfp->bf_levels++;
				temp /= bfp->bf_filesperdir;
			    }

			    // temp debug
			    CDEBUG((D_LOOKATTR, "bigfile converted data: "
				    "size %q hunkbytes %d fperdir %d "
				    "files %d levels %d dirdgts %d\n",
				    bfp->bf_size.QuadPart, bfp->bf_hunkbytes,
				    bfp->bf_filesperdir, bfp->bf_files,
				    bfp->bf_levels, bfp->bf_dirdigits));
			}
		    } else
			CDEBUG((D_LOOKATTR, "bigfile data: Read error\n"));

		    // Clean up now
		    coda_release(bfifh);
		    IoFreeMdl(pmdl);
		    coda_free(TempData);
		}
	    }
#endif
	    
	}

	// Set the FileAttributes if requested.

	if (FileAttributes != NULL) {
	    __try {
		// Return the attributes.
		RtlCopyMemory(FileAttributes, &inode->c_fattr,
			      sizeof(FS_FILE_ATTRIBUTES));
	    }
	    except (EXCEPTION_EXECUTE_HANDLER) {
		CDEBUG((D_WRAPPER,
			"coda_internal_getattr: can't copy attributes.\n"));
		return STATUS_INVALID_USER_BUFFER;
	    }
#ifdef _BIG_FILE_
	    if (inode->c_big_filep) {
		CDEBUG((D_LOOKATTR,
			"coda_internal_getattr: returning big file attrs.\n"));
		// Return the fake information
		FileAttributes->Attributes = FILE_ATTRIBUTE_READONLY ;
		FileAttributes->LastByteWritten = inode->c_big_filep->bf_size;
		FileAttributes->ValidDataLength = inode->c_big_filep->bf_size;
		FileAttributes->DiskLength = inode->c_big_filep->bf_size;
	    }
#endif
	    CDEBUG ((D_LOOKATTR, "coda_internal_getattr: LastByte %q, "
		     "ValidData %q, DiskLen %q\n",
		     FileAttributes->LastByteWritten.QuadPart,
		     FileAttributes->ValidDataLength.QuadPart,
		     FileAttributes->DiskLength.QuadPart));
	    
	}
    } 

    finally {
	UnLockInode(inode);
    }

    CDEBUG ((D_LOOKATTR, "coda_internal_getattr: finished with %s\n",
	     coda_f2s(&inode->c_fid)));
    return STATUS_SUCCESS;
}

//
// coda_getattr
//
NTSTATUS coda_getattr(IN FS_FILE_HANDLE FileHandle,	
			       IN OUT PFS_FILE_ATTRIBUTES FileAttributes)
{
    coda_filehandle *fh = (coda_filehandle *)FileHandle;
    coda_inode * inode = fh->fh_inode;
    NTSTATUS rc;

    CDEBUG((D_LOOKATTR, "coda_getattr: Started on %s. (fh %d)\n",
	    inode->c_name, fh->fh_unique));

    rc = coda_internal_getattr (inode, fh->fh_uid, TRUE, FileAttributes);

    CDEBUG((D_LOOKATTR, "coda_getattr: Exiting. (fh %d)\n",
	    fh->fh_unique));

    return rc;
}

//
// Support for coda_setattr 
//

static void coda_FATTR2vattr(PFS_FILE_ATTRIBUTES FileAttributes,
			     struct coda_vattr *Attributes)
{
    //
    // Convert times ... which is about all we can do to the attrs!
    //

    if (!RtlTimeToSecondsSince1970(&FileAttributes->LastAccessTime,
				   &Attributes->va_atime.ts_sec)) {
	CDEBUG ((D_LOOKATTR,
		 "coda_FATTR2vattr: Failed to convert LastAccessTime\n"));
	Attributes->va_atime.ts_sec = 0;
	Attributes->va_atime.ts_nsec = 0;
    }

    if (!RtlTimeToSecondsSince1970(&FileAttributes->LastModifiedTime,
				   &Attributes->va_mtime.ts_sec)) {
	CDEBUG ((D_LOOKATTR,
		 "coda_FATTR2vattr: Failed to convert LastModifiedTime\n"));
	Attributes->va_mtime.ts_sec = 0;
	Attributes->va_mtime.ts_nsec = 0;
    }
    
    return;
    
}

//
// coda_setattr
//

NTSTATUS coda_setattr(IN FS_FILE_HANDLE FileHandle,
		      IN OUT PFS_FILE_ATTRIBUTES FileAttributes)
{

    NTSTATUS rc;
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode * inode = fh->fh_inode;
    coda_ntvolume *ntvol;
    struct coda_vattr Attributes;
    
    CDEBUG((D_LOOKATTR, "coda_setattr: Started on %s (fh %d).\n",
	    inode->c_name, fh->fh_unique));
    
    
    //
    // Do some sanity checking to be sure that the handle that
    // was passed in is really a valid file handle.
    //
    if (! CODA_ISOURS(inode) )  {
	CDEBUG((D_WRAPPER, "coda_setattr: NOT A FILE!\n"));
	return(STATUS_INVALID_PARAMETER_1);
    }

    LockInode (inode);
    
    // Set the fields ...
    coda_FATTR2vattr(FileAttributes, &Attributes);
    Attributes.va_ctime.ts_nsec = -1;
    Attributes.va_ctime.ts_sec = -1;
    Attributes.va_type = C_VNON;
    if (FileAttributes->Attributes & FILE_ATTRIBUTE_READONLY)
	Attributes.va_mode = C_M_READ;
    else
	Attributes.va_mode = -1;
    Attributes.va_nlink = -1;
    Attributes.va_uid = -1;
    Attributes.va_gid = -1;
    Attributes.va_fileid = -1;
    Attributes.va_size = -1;
    Attributes.va_blocksize = -1;
    Attributes.va_gen = -1;
    Attributes.va_flags = -1;
    Attributes.va_rdev = -1;
    Attributes.va_bytes = -1;
    Attributes.va_filerev = -1;
    
    rc = venus_setattr(&inode->c_fid, &Attributes, fh->fh_uid);
    
    if( rc ) {
	CDEBUG((D_LOOKATTR, "coda_setattr: venus_setattr %s "
		"returns error %d\n", coda_f2s(&inode->c_fid), rc));
	UnLockInode (inode);
	return(rc);
    }
    
    /* Just update the Attar  */
    inode->c_fattr.LastAccessTime = FileAttributes->LastAccessTime;
    inode->c_fattr.LastModifiedTime = FileAttributes->LastModifiedTime;

    UnLockInode (inode);
    
    EXIT(D_LOOKATTR, "coda_setattr");
    return STATUS_SUCCESS;
}


//
// coda_lookup -- implements the FS_LOOKUP wrapper call
//
NTSTATUS coda_internal_lookup(IN coda_inode *dirinode,
			      IN PCHAR FileName,
			      IN int NameLen,
			      IN BOOLEAN CaseInsensitive,
			      IN int Uid,
			      IN BOOLEAN CheckForBF,
			      IN BOOLEAN FindSymLinks,
			      OUT coda_inode **fileinode,
			      OUT coda_filehandle **pfh)
{
    NTSTATUS Status;
    
    struct coda_vattr Attributes;

    coda_inode *inode = NULL;
    CodaFid fid;
    int type;

    char BaseName[CODA_MAXNAMLEN];
    int BaseNameLen = 0;
    
    //
    // Null the return parameters, in case we bail-out early
    //
    *fileinode = NULL;
    
    //
    // Looks like the input parameters are OK.  Let's do the
    // lookup...
    //
    CDEBUG((D_LOOKATTR, "coda_internal_lookup: Lookup of %s in %s%s\n",
	   FileName, coda_f2s(&dirinode->c_fid),
	   (CaseInsensitive?" Case Insensitive":"")));

    // Quick check for autorun.inf ...
    if ((dirinode == root_fh->fh_inode)
	&& (strcasecmp(FileName, "autorun.inf") == 0)) {
	CDEBUG((D_LOOKATTR, "coda_internal_lookup: no autorun.inf\n"));
	return STATUS_OBJECT_NAME_NOT_FOUND;
    }

    LockInode(dirinode);

    // Fisrt, check cache

    GrabLock();
    inode = inode_find_name(dirinode, FileName, CaseInsensitive);
    DropLock();

    if ( !inode && (CaseInsensitive ?
		    strcasetail (FileName, ".lnk") :
		    strtail (FileName, ".lnk")) ) {
	
	// Get the base name of a .lnk file
	BaseNameLen = NameLen - 4;
	memcpy (BaseName, FileName, BaseNameLen);
	BaseName[BaseNameLen] = 0;
	
	// Check cache for the base name
	GrabLock();
	inode = inode_find_name(dirinode, BaseName, CaseInsensitive);
	DropLock();
	
    }
    
    if (!inode) {
	CDEBUG((D_LOOKATTR, "coda_internal_lookup: "
	       "cache miss, going to Venus\n"));

	Status = venus_lookup(&dirinode->c_fid, (const char *)FileName, 
			      NameLen, &type, &fid, Uid);
	
	if (Status && BaseNameLen > 0) {
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: %s not found, "
		    "trying %s\n", FileName, BaseName));
	    
	    Status = venus_lookup(&dirinode->c_fid, (const char *)BaseName, 
				  NameLen, &type, &fid, Uid);
	    if (!Status && ((type != C_VLNK) && (type >= C_VNON))) { // ZZZZZZ
		CDEBUG((D_LOOKATTR, "coda_internal_lookup: %s not symlink\n",
			BaseName));
		Status = STATUS_OBJECT_NAME_NOT_FOUND;
	    }
	} else
	    BaseNameLen = 0;
	
	if (Status) {
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: "
		   "venus_lookup failed. Status = 0x%x\n",
		   Status));
	    UnLockInode(dirinode);
	    return STATUS_OBJECT_NAME_NOT_FOUND;
	}

	GrabLock();
	if (BaseNameLen > 0)
	    inode =  inode_new(&fid, dirinode, BaseName);
	else
	    inode =  inode_new(&fid, dirinode, FileName);
	DropLock();
	
	if(!inode) {
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: inode_new failed!\n"));
	    UnLockInode(dirinode);
	    return STATUS_INVALID_PARAMETER_1;
	}

	// Set the inode fstype
	switch (type) {
	  case C_VREG:
	      inode->c_fstype = OSRFS_TYPE_FILE;
	      CDEBUG ((D_LOOKATTR, "coda_internal_lookup: "
		       "venus returns type file\n"));
	      break;
	  case C_VDIR:
	      inode->c_fstype = OSRFS_TYPE_DIRECTORY;
	      CDEBUG ((D_LOOKATTR, "coda_internal_lookup: "
		       "venus returns type directory\n"));
	      break;
	  case C_VLNK:
	      inode->c_fstype = OSRFS_TYPE_FILE;
	      inode->c_lnk_state = SLNK_NORMAL;
	      CDEBUG ((D_LOOKATTR, "coda_internal_lookup: "
		       "venus returns type symlink\n"));
	      break;
	  default:
	      inode->c_fstype = OSRFS_TYPE_OTHER;
	      CDEBUG ((D_LOOKATTR, "coda_internal_lookup: "
		       "venus returns type other (%d)!!\n",
		       type));
	}
    } else {
	CDEBUG((D_LOOKATTR, "coda_internal_lookup: cache hit (%s).\n",
		inode->c_name));
    }

    if (!inode_hasflag (inode, I_VATTR)) {
	Status = coda_internal_getattr(inode, Uid, CheckForBF, NULL);
	if ( !NT_SUCCESS(Status) ) {
	    CDEBUG((D_LOOKATTR, "coda_internal_lookup: "
		   "Cannot get attributes of inode %s "
		   "during lookup: %d\n",
		   coda_f2s(&inode->c_fid), Status));
	    GrabLock();
	    inode_release (inode);	
	    DropLock();
	    UnLockInode(dirinode);
	    return STATUS_OBJECT_NAME_NOT_FOUND;
	}
    }

    UnLockInode(dirinode);

    *fileinode = inode;

    // Symbolic link "support"
    if ( !FindSymLinks && CODA_ISSYMLINK(inode) && (BaseNameLen == 0) )
	    return STATUS_OBJECT_NAME_NOT_FOUND;

#ifdef _BIG_FILE_
    if (inode->c_big_filep)
	CDEBUG((D_LOOKATTR, "coda_internal_lookup: inode is a big file\n"));
#endif    

    // Get and fill in a file handle if one sent into the routine.
    if (pfh != NULL) {
	if (inode == coda_nt_volume.v_ctlinode) {
	    *pfh = coda_nt_volume.v_ctlfh;

	    CDEBUG ((D_LOOKATTR, "coda_internal_lookup: "
		     "returning control fh %d\n", (*pfh)->fh_unique));
	} else {

	    GrabLock();
	    // Note: get_cfh produces a unique fh for each uid.
	    *pfh = get_cfh(inode, dirinode, Uid);
	    DropLock();
	    
	    if (!*pfh) {
		CDEBUG ((D_LOOKATTR, "coda_internal lookup: "
			 "could not allocate a file handle.\n"));
		inode_release(inode);
		return STATUS_INSUFFICIENT_RESOURCES;
	    }
	    
	    CDEBUG ((D_LOOKATTR, "coda_internal_lookup: returning fh %d\n",
		     (*pfh)->fh_unique));
	}
    }
    
    CDEBUG((D_LOOKATTR, "coda_internal_lookup: returning success\n"));
    return STATUS_SUCCESS;
}


NTSTATUS coda_lookup(IN FS_FILE_HANDLE DirectoryHandle,
		     IN PWCHAR FileName,
		     IN BOOLEAN CaseInsensitive,
		     OUT PFS_FILE_HANDLE FileHandle,
		     OUT PULONG Type)
     
{
    coda_filehandle *dfh = (coda_filehandle *)DirectoryHandle;
    coda_filehandle *nfh;   
    ANSI_STRING AnsiString;
    UNICODE_STRING UnicodeString;
    CHAR NormalFileName[CODA_MAXNAMLEN];

    coda_inode *dirinode = dfh->fh_inode;
    coda_inode *inode;
    int namelen;
    int linklook = FALSE;
    int rc;
    int Uid = GetNtUID ();

    CDEBUG ((D_LOOKATTR, "coda_lookup: dir fh %d\n", dfh->fh_unique));

    //
    // First, do some sanity checking to be sure that the handle
    // that was passed in is really a valid directory handle.
    // For us, this means that it is a pointer to an inode. (XXX?)
    //
    if (!CODA_ISOURS(dirinode) ) {
	return(STATUS_INVALID_PARAMETER_1);
    }
    
    if (!CODA_ISDIR(dirinode)) {
	return(STATUS_NOT_A_DIRECTORY);
    }

    //
    // Next, convert the wide-character string to a "normal" null
    // terminated c-string. One would think this would be easier.
    //
    RtlInitUnicodeString(&UnicodeString, FileName);
    RtlUnicodeStringToAnsiString(&AnsiString, &UnicodeString, TRUE);
	
    if (AnsiString.Length > CODA_MAXNAMLEN) {
	return(STATUS_INVALID_PARAMETER_2);
    }

    namelen = AnsiString.Length;
    memcpy(NormalFileName, AnsiString.Buffer, namelen);
    NormalFileName[namelen] = 0x00;
    RtlFreeAnsiString(&AnsiString);

    CDEBUG ((D_LOOKATTR, "coda_lookup: file name '%s'\n", NormalFileName));

    rc = coda_internal_lookup (dfh->fh_inode, NormalFileName, namelen,
			       CaseInsensitive, Uid, TRUE, FALSE,
			       &inode, &nfh);

    if ( NT_SUCCESS (rc) ) {

	*FileHandle = (FS_FILE_HANDLE) nfh;
#ifdef _BIG_FILE_
	if ( inode->c_big_filep ) {
	    CDEBUG((D_LOOKATTR, "coda_lookup: returning file type "
		    "for big file\n"));
	    *Type = OSRFS_TYPE_FILE;
	} else
#endif
	    *Type = inode->c_fstype;
    }
    
    // Checked for zapped dirs ...

    if ( inode_hasflag(dirinode, I_ZAPPED) ) {
 	coda_filehandle *fh;
	inode_unsetflag (dirinode, I_ZAPPED);
	purge_fh_list (dirinode);
    }    
    
    EXIT (D_LOOKATTR, "coda_lookup");
    return rc;
}

//
// Path lookup  -- XXX make similar to coda_internal_lookup
//

NTSTATUS coda_internal_path_lookup(IN coda_inode *dirinode,
				   IN PCHAR FilePath,
				   IN BOOLEAN CaseInsensitive,
				   IN int Uid,
				   OUT coda_inode **fileinode)
{
    NTSTATUS Status;
    
    CHAR NormalFilePath[CODA_MAXNAMLEN];
    CHAR *FileName;
    CHAR *RemainPath;
    CHAR *temp;
    
    coda_inode *inode = NULL;
    int namelen;

    ENTRY (D_PIOCTL, "coda_internal_path_lookup"); 

    // If we fall through everything, there was "no" file name,
    // so we return the coda root directory.
    Status = STATUS_SUCCESS;
    inode = coda_nt_volume.v_rootinode; 

    // Get the string in the local space for modification.

    if (strlen(FilePath) > CODA_MAXNAMLEN) {
	return(STATUS_INVALID_PARAMETER_2);
    }
    strcpy(NormalFilePath, FilePath);

    // Make sure all file separators are /.
    temp = NormalFilePath;
    while (*temp) {
	if (*temp == '\\')
	    *temp = '/';
	temp++;
    }

    // Walk the file name ...
    RemainPath = NormalFilePath;
    while (*RemainPath == '/')
	RemainPath++;
    while (*RemainPath) {
	FileName = RemainPath;
	while (*RemainPath && (*RemainPath != '/'))
	    RemainPath++;
	if (*RemainPath == '/') {
	    *RemainPath = 0;
	    namelen = RemainPath - FileName;
	    RemainPath++;
	    while (*RemainPath == '/')
		RemainPath++;
	}

	CDEBUG( (D_PIOCTL, "coda_internal_path_lookup: "
		"FileName = '%s' (len %d), RemainPath= '%s'\n", FileName,
		 namelen, RemainPath)); 

	Status = coda_internal_lookup (dirinode, FileName, namelen,
				       CaseInsensitive,
				       Uid, TRUE, FALSE, &inode, NULL);

	if (Status != STATUS_SUCCESS) {
	    CDEBUG( (D_PIOCTL,
		    "coda_internal_path_lookup: returning error %d\n",
		     Status));
	    return Status;
	}
	dirinode = inode;
    }

    *fileinode = inode;

    EXIT (D_PIOCTL, "coda_internal_path_lookup"); 

    return Status;
}

#ifdef _BIG_FILE_

// Support routines for coda_big_file

NTSTATUS vget_hunk ( IN coda_filehandle *fh,
		     IN ULONG hunknum,
		     IN bf_hunk *hunk,
		     IN coda_big_file *bf)
{
    NTSTATUS status = STATUS_SUCCESS;
    LONG temp;
    ULONG temphunk = hunknum;
    coda_inode *inode;
    coda_inode *dirnode;

    CDEBUG ((D_LOOKATTR, "vget_hunk: getting hunk %d (filesperdir %d)\n",
	     hunknum, bf->bf_filesperdir));

    if (hunknum >= bf->bf_files) {
	CDEBUG ((D_LOOKATTR, "vget_hunk: EOF\n"));
	return STATUS_END_OF_FILE;
    }

    if (bf->bf_filesperdir == 0) {
	CDEBUG ((D_LOOKATTR, "vget_hunk: bf_filesperdir is zero.\n"));
	return STATUS_INSUFFICIENT_RESOURCES;
    }

    /* Make dir and file names */
    hunk->hu_num = hunknum;
    for (temp = 0 ; temp <= bf->bf_levels; temp++) {
	DirName(bf->bf_dirdigits, temphunk % bf->bf_filesperdir,
		hunk->hu_name[temp]);
	temphunk = temphunk / bf->bf_filesperdir;
	CDEBUG ((D_LOOKATTR, "vget_hunk: hu_name[%d]->%s\n",
		 temp, hunk->hu_name[temp] ));
    }

    /* get the file handles. */
    dirnode = fh->fh_inode;
    for (temp = bf->bf_levels; temp >= 0; temp--) {
	status = coda_internal_lookup (dirnode, hunk->hu_name[temp],
				       bf->bf_dirdigits,
				       TRUE, fh->fh_uid, FALSE, FALSE, &inode,
				       &hunk->hu_fh[temp]);
	if (!NT_SUCCESS(status)) {
	    CDEBUG((D_LOOKATTR, "vget_hunk: could not get entry %d\n", temp));
	    hunk->hu_num = -1;
	    return status;
	}
	dirnode = inode;
    }	
    
    /* vget the file -- filehandle in hunk->hu_fh[0] for inode. */
    {
	CodaFid resfid;
	int type;
	int rc;

	rc = venus_vget(&inode->c_fid, &resfid, &type, fh->fh_uid);

	if (rc) 
	    CDEBUG((D_LOOKATTR, "vget_hunk: venus_vget failed!\n"));
    } 

    CDEBUG ((D_LOOKATTR, "vget_hunk: returning %d\n", status));
    return status;
}


// coda_read_bigfile
//
//   This routine reads from the directory tree under the special
//   big file directory.  It uses coda_read to get the job done.
//   It is also called from coda_read!
//
//   Arguments are identical to coda_read (see below)
//

NTSTATUS coda_read_bigfile (IN coda_filehandle *fh,
			    IN LARGE_INTEGER FileOffset,
			    IN PMDL Buffer,
			    IN ULONG ReadSize,
			    OUT PULONG BytesRead)
{
    NTSTATUS status;
    int temp;
    bf_hunk temphunk;
    coda_big_file *bf_filep = fh->fh_inode->c_big_filep;
    
    ULONG hunknum; 
    ULONG newReadSize;

    CDEBUG((D_WRAPPER, "coda_read_bigfile: Reading %s "
	    "(fh %d, offset %q, size %d).\n",  fh->fh_inode->c_name,
	    fh->fh_unique, FileOffset.QuadPart, ReadSize));

    if (bf_filep->bf_hunkbytes == 0) {
	CDEBUG((D_WRAPPER, "coda_read_bigfile: hunkbytes is zero.\n"));
	return STATUS_INSUFFICIENT_RESOURCES;
    }

    if (FileOffset.QuadPart >= bf_filep->bf_size.QuadPart) {
	CDEBUG((D_WRAPPER, "coda_read_bigfile: EOF.\n"));
	return STATUS_END_OF_FILE;
    }

    hunknum = (long) (FileOffset.QuadPart / bf_filep->bf_hunkbytes);
    newReadSize = (long) (((LONGLONG) hunknum+1)*(bf_filep->bf_hunkbytes)
				    - FileOffset.QuadPart);
    if (ReadSize < newReadSize)
	newReadSize = ReadSize;

    CDEBUG((D_WRAPPER, "coda_read_bigfile: hunk %d, newReadSize %d.\n",
	    hunknum, newReadSize));

    /* Quick stuff check */
    if (hunknum == bf_filep->bf_curr.hu_num)
	if (bf_filep->bf_curr.hu_fh[0]) {
	    FileOffset.QuadPart -= hunknum*bf_filep->bf_hunkbytes;
	    return coda_read((FS_FILE_HANDLE)bf_filep->bf_curr.hu_fh[0],
			     FileOffset, Buffer, newReadSize, BytesRead);
	} else {
	    CDEBUG ((D_WRAPPER, "coda_read_bigfile: curr.hu_fh not valid!\n"));
	    return STATUS_INSUFFICIENT_RESOURCES;
	}

    /* Not current, release current */
    for (temp = 0; temp < 3 ; temp++)
	if (bf_filep->bf_curr.hu_fh[temp]) {
	    if (bf_filep->bf_curr.hu_fh[temp] !=
		bf_filep->bf_next.hu_fh[temp]) {
		inode_setflag (bf_filep->bf_curr.hu_fh[temp]->fh_inode,
			       I_PURGE);
		coda_release (bf_filep->bf_curr.hu_fh[temp]);
	    }
	    bf_filep->bf_curr.hu_fh[temp] = NULL;
	}
    bf_filep->bf_curr.hu_num = -1;

    /* Is it the next one? */
    if (hunknum == bf_filep->bf_next.hu_num) {

	/* yes, move next to current. */
	bf_filep->bf_curr = bf_filep->bf_next;
	for (temp = 0; temp < 3 ; temp++)
	    bf_filep->bf_next.hu_fh[temp] = NULL;
	bf_filep->bf_next.hu_num = -1;

    } else {

	/* Completely new */
	for (temp = 0; temp < 3 ; temp++)
	    if (bf_filep->bf_next.hu_fh[temp]) {
		inode_setflag (bf_filep->bf_next.hu_fh[temp]->fh_inode,
			       I_PURGE);
		coda_release (bf_filep->bf_next.hu_fh[temp]);
		bf_filep->bf_next.hu_fh[temp] = NULL;
	    }
	bf_filep->bf_next.hu_num = -1;

	status = vget_hunk (fh, hunknum, &bf_filep->bf_curr, bf_filep);
	if (!NT_SUCCESS(status)) {
	    CDEBUG((D_WRAPPER, "coda_read_bigfile: vget_hunk %d failed.\n",
		    hunknum));
	    return status;
	}
    }

    // Adjust FileOffset and read ...
    FileOffset.QuadPart -= hunknum*bf_filep->bf_hunkbytes;
    status = coda_read((FS_FILE_HANDLE)bf_filep->bf_curr.hu_fh[0],
		       FileOffset, Buffer, newReadSize, BytesRead); 

    if (!NT_SUCCESS(status))
	return status;
    
    if (!NT_SUCCESS(vget_hunk (fh, hunknum+1, &bf_filep->bf_next, bf_filep))) {
	for (temp = 0; temp < 3 ; temp++)
	    if (bf_filep->bf_next.hu_fh[temp]) {
		inode_setflag (bf_filep->bf_next.hu_fh[temp]->fh_inode,
			       I_PURGE);
		coda_release (bf_filep->bf_next.hu_fh[temp]);
		bf_filep->bf_next.hu_fh[temp] = NULL;
	    }
	bf_filep->bf_next.hu_num = -1;
    }

    EXIT (D_WRAPPER, "coda_read_bigfile:");
    return status;
}

#endif

//
// coda_read
//
//   This routine reads data from the container file.
//
// Inputs:
//   FileHandle - the file being read
//   FileOffset - the first byte of the read
//   Buffer - the buffer where the data should be stored
//   ReadSize - the number of bytes to be read.
//
// Outputs:
//   BytesRead - the number of bytes actually read
//
// Returns:
//   SUCCESS - the operation completed successfully
//
// Notes:
//   None.
//

NTSTATUS coda_read(IN FS_FILE_HANDLE FileHandle,
		   IN LARGE_INTEGER FileOffset,
		   IN PMDL Buffer,
		   IN ULONG ReadSize,
		   OUT PULONG BytesRead)
{
    PIRP irp;
    KEVENT event;
    PIO_STACK_LOCATION ioStackLocation;
    PDEVICE_OBJECT fsdDevice;
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode * inode = fh->fh_inode;
    IO_STATUS_BLOCK iosb;
    
    ENTRY(D_WRAPPER, "coda_read");
    
    //
    // Sanity checking
    //
    if ( !CODA_ISOURS(inode) || fh == coda_nt_volume.v_ctlfh) {
	CDEBUG((D_WRAPPER, "coda_read: Bad file %s\n", inode->c_name));
	return STATUS_INVALID_PARAMETER_1;
    }
    
    CDEBUG((D_WRAPPER, "coda_read: Reading %s (fh %d, offset %q, size %d).\n",
	    inode->c_name, fh->fh_unique, FileOffset.QuadPart, ReadSize));

#ifdef _BIG_FILE_
    if (inode->c_big_filep) {
	return coda_read_bigfile (fh, FileOffset, Buffer, ReadSize, BytesRead);
    }
#endif

    if (CODA_ISSYMLINK(inode)) {
	char *BufPtr;
	ULONG bytescopied = 0;

	CDEBUG((D_WRAPPER, "coda_read: Read Symbolic Link.\n"));

	if (inode->c_lnk_file) {
	    if ((inode->c_lnk_len > FileOffset.QuadPart) &&
		(inode->c_lnk_state == SLNK_NORMAL)) {
		BufPtr = MmGetSystemAddressForMdl(Buffer);
		bytescopied = inode->c_lnk_len-(long)FileOffset.QuadPart;
		if (bytescopied > ReadSize)
		    bytescopied = ReadSize;
		memcpy(BufPtr, inode->c_lnk_file+FileOffset.QuadPart,
		       bytescopied);
	    }
	} else {
	    CDEBUG((D_WRAPPER, "coda_read: No file buffer to 'read'\n"));
	}

	*BytesRead = bytescopied;
	CDEBUG((D_WRAPPER, "coda_read: Read Symbolic Link, bytes=%d.\n",
		bytescopied));
	return STATUS_SUCCESS;
    }

    if ( coda_get_fileobject(fh, C_O_READ) ) {
	CDEBUG((D_WRAPPER, "coda_read: coda_get_fileobject error.\n"));
	return STATUS_INVALID_PARAMETER_1;
    }
    
    //
    // Figure out where to send the IRP
    //
    fsdDevice = IoGetRelatedDeviceObject(inode->c_object);
    
    //
    // Set up our event for use.
    //
    KeInitializeEvent(&event, NotificationEvent, FALSE);
    
    //
    // Allocate and build the IRP we'll be sending to the FSD.
    //    
    irp = IoAllocateIrp(fsdDevice->StackSize, FALSE);    
    if (!irp) {        
	//
	// Allocation failed, presumably due to memory allocation
	// failure.
	//        
	return STATUS_INSUFFICIENT_RESOURCES;        
    }
    
    irp->MdlAddress = Buffer;    
    irp->UserEvent = &event;    
    irp->UserIosb = &iosb;    
    irp->Tail.Overlay.Thread = PsGetCurrentThread();    
    irp->Tail.Overlay.OriginalFileObject= inode->c_object;
    irp->RequestorMode = KernelMode;
    
    //
    // Indicate that this is a READ operation.
    //    
    irp->Flags = IRP_READ_OPERATION | IRP_NOCACHE;
    irp->UserBuffer = MmGetMdlVirtualAddress(Buffer);    
    
    //
    // Set up the next I/O stack location.  These are the parameters
    // that will be passed to the underlying driver.
    //
    
    ioStackLocation = IoGetNextIrpStackLocation(irp);    
    ioStackLocation->MajorFunction = IRP_MJ_READ;    
    ioStackLocation->MinorFunction = 0;    
    ioStackLocation->DeviceObject = fsdDevice;    
    ioStackLocation->FileObject = inode->c_object;
    ioStackLocation->Parameters.Read.Length = ReadSize;
    ioStackLocation->Parameters.Read.ByteOffset = FileOffset;
    
    //
    // Use the generic completion routine.  This ensures two things:
    //  (1) the MDL we pass won't get freed;
    //  (2) we get I/O completion without the cost.
    //
    IoSetCompletionRoutine(irp, NullDiscardIrpCompletion, 0, TRUE, TRUE,
			   TRUE);
    
    
    //
    //  I/O operation is built.  Pass it to the underlying file system.
    //
    (void) IoCallDriver(fsdDevice, irp);
    
    
    //
    // Wait for the I/O to complete.
    //
    KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0);
    
    
    //
    // Copy results from io status block.
    //
    *BytesRead = iosb.Information;
    
    
    //
    // Done.
    //
    CDEBUG ((D_WRAPPER, "coda_read: exit: %d bytes read, status %x\n",
	     *BytesRead, iosb.Status));
    return iosb.Status;
}

// coda_set_length -- It appears this must be called before
// any call to coda_write ... to make sure there is enough
// room in the file for the write.

NTSTATUS coda_set_length(IN FS_FILE_HANDLE FileHandle,
			 IN LARGE_INTEGER AllocationSize,
			 IN LARGE_INTEGER DataSize)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle; 
    coda_inode * inode = fh->fh_inode;
    struct coda_vattr f_attr;
    int rc;

    CDEBUG ((D_WRAPPER, "coda_set_length: Allocate %q, DataSize %q\n",
	     AllocationSize.QuadPart, DataSize.QuadPart));

    if ( !CODA_ISOURS(inode) || fh == coda_nt_volume.v_ctlfh ) {
	CDEBUG((D_WRAPPER, "coda_set_length: Bad file %s\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    CDEBUG((D_WRAPPER, "coda_set_length: Setting %s. (fh %d)\n",
	    inode->c_name, fh->fh_unique));

    if (CODA_ISSYMLINK(inode)) {
	// Change malloc size
	CDEBUG((D_WRAPPER, "coda_set_length: symlink setsize.\n"));
	if (inode->c_lnk_file != NULL)
	    coda_free(inode->c_lnk_file);
	inode->c_lnk_file = (char *)coda_malloc((int)DataSize.QuadPart);
	if (inode->c_lnk_file)
	    inode->c_lnk_len = (int)DataSize.QuadPart;
	else {
	    CDEBUG((D_WRAPPER, "coda_set_length: symlink malloc error.\n"));
	    return STATUS_INSUFFICIENT_RESOURCES;
	}
	inode->c_lnk_state = SLNK_READY;  // Ready to be written

    } else if ( coda_get_fileobject(fh, C_O_WRITE) ) {
	CDEBUG((D_WRAPPER, "coda_set_length: coda_get_fileobject error.\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    LockInode(inode);

    /* Update Local copy! */
    inode->c_fattr.ValidDataLength = DataSize;
    inode->c_fattr.DiskLength = AllocationSize;
    rc = STATUS_SUCCESS;
    
    UnLockInode(inode);
    CDEBUG ((D_WRAPPER, "coda_set_length: Exiting. (fh %d)\n", fh->fh_unique));
    return rc;
}


//
// coda_write
//
//   This routine writes data to the backing store
//
// Inputs:
//   FileHandle - the file being read
//   FileOffset - the first byte of the read
//   Buffer - the buffer where the data should be stored
//   WriteSize - the number of bytes to be read.
//
// Outputs:
//   BytesWritten - the number of bytes actually written.
//
// Returns:
//   SUCCESS - the operation was successful.
//
// Notes:
//   None.
//
NTSTATUS coda_write(IN FS_FILE_HANDLE FileHandle,
		    IN LARGE_INTEGER FileOffset,
		    IN PMDL Buffer,
		    IN ULONG WriteSize,
		    OUT PULONG BytesWritten)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode * inode = fh->fh_inode;

    PIRP irp;
    KEVENT event;
    PIO_STACK_LOCATION ioStackLocation;
    PDEVICE_OBJECT fsdDevice;
    IO_STATUS_BLOCK iosb;
    
  
    //
    // Sanity checking
    //
    if ( !CODA_ISOURS(inode)  || fh == coda_nt_volume.v_ctlfh ) {
	CDEBUG ((D_WRAPPER, "coda_write: Bad file\n"));
	return STATUS_INVALID_PARAMETER_1;
    }

    CDEBUG ((D_WRAPPER,
	     "coda_write: Writing %s (fh %d) Offset %q WriteSize %d.\n",
	     inode->c_name, fh->fh_unique, FileOffset.QuadPart, WriteSize ));

    
    if (CODA_ISSYMLINK(inode)) {
	// Symlink write ... get data.
	char *BufAdr = MmGetSystemAddressForMdl(Buffer);
	if (FileOffset.QuadPart+WriteSize > (LONGLONG)inode->c_lnk_len) {
	    CDEBUG((D_WRAPPER, "coda_write: link data too big (max %d).\n",
		    inode->c_lnk_len));
	    return STATUS_INVALID_PARAMETER_1;
	}
	memcpy(inode->c_lnk_file+(int)FileOffset.QuadPart,
	       BufAdr, WriteSize);
	*BytesWritten = WriteSize;

	// Check for full link written ...
	if (FileOffset.QuadPart+WriteSize == (LONGLONG)inode->c_lnk_len
	    && inode->c_lnk_state == SLNK_READY) {

	    // We should only get here if we are creating a new .lnk file!
	    lnk_head *newlnk = (lnk_head *)inode->c_lnk_file;
	    char *lnkdata = inode->c_lnk_file + sizeof(lnk_head);
	    short dsize;
	    int ix;
	    int issymlink;
	    int rc;

	    CDEBUG ((D_WRAPPER, "coda_write: symlink file complete.\n"));

	    issymlink = newlnk->lk_magic == lnk_header.lk_magic;
	    for (ix = 0; issymlink && ix < 16; ix++)
		issymlink = newlnk->lk_guid[ix] == lnk_header.lk_guid[ix];
	    
	    // Make sure it just has a description string and a relative path
	    // Ignore shell item id list.  Any other element makes it look
	    // like a regular file.
	    issymlink = issymlink && ((newlnk->lk_flags & 0x7e) == 0x0c);
	    if (issymlink) {
		int inamelen = strlen(inode->c_name);
		int type;
		CDEBUG((D_WRAPPER,
			"coda_write: storing .lnk file as symlink.\n"));
		// Extract description string and send to venus
		// Leave buffer in place
		if (newlnk->lk_flags & 0x1) { // Shell item present
		    dsize = *((short *)lnkdata);
		    lnkdata += dsize + 2;
		}
		//  This must be the description string given earlier checks.
		dsize = *((short *)lnkdata);
		lnkdata += 2;  // Now points to description
		rc = venus_remove (&inode->c_parent->c_fid, inode->c_name,
				   inamelen, fh->fh_uid);
		if (rc != 0) {
		    CDEBUG ((D_WRAPPER,
			     "coda_write: can't remove tmp symlink.\n"));
		}
		// Change the name. (remove .lnk)
		inamelen -= 4;
		inode->c_name[inamelen] = 0;

		rc = venus_symlink (&inode->c_parent->c_fid, inode->c_name,
				    inamelen, lnkdata, dsize, fh->fh_uid);
		if (rc != 0) {
		    CDEBUG ((D_WRAPPER,
			     "coda_write: error making symlink.\n"));
		}
		rc = venus_lookup (&inode->c_parent->c_fid, inode->c_name,
				   inamelen, &type, &inode->c_fid, fh->fh_uid);
		if (rc != 0) {
		    CDEBUG ((D_WRAPPER,
			     "coda_write: lookup on new symlink failed.\n"));
		}
		inode->c_fattr.Attributes = FILE_ATTRIBUTE_READONLY;
		inode->c_lnk_state = SLNK_NORMAL;
		inode_rename(inode, inode->c_parent, inode->c_name);
		purge_fh_list (inode);
		purge_fh_list (inode->c_parent);
	    } else {
		// This doesn't look like a cygwin .lnk file, do
		// the real write to the file system.
		LARGE_INTEGER Offset;
		ULONG DidWrite;
		PMDL pmdl;
		NTSTATUS Wstat;

		CDEBUG ((D_LOOKATTR,
			 "coda_write: doing real write to .lnk file.\n"));

		inode->c_lnk_state = SLNK_NONE;
		Offset.QuadPart = 0;
		pmdl=IoAllocateMdl (inode->c_lnk_file, inode->c_lnk_len,
				    FALSE, FALSE, NULL);
		if (!pmdl) {
		    CDEBUG ((D_LOOKATTR,
			     "coda_write: failed to allocate Mdl.\n"));
		    return STATUS_INSUFFICIENT_RESOURCES;
		}

		MmBuildMdlForNonPagedPool (pmdl);		    

		// Recursive Call! 
		Wstat = coda_write (fh, Offset, pmdl, inode->c_lnk_len,
				    &DidWrite);

		IoFreeMdl(pmdl);
		
		if (!NT_SUCCESS(Wstat)) {
		    CDEBUG ((D_WRAPPER,
			     "coda_write: recursive write error %d\n",
			     Wstat));
		    return Wstat;
		}

		// Calculate a good *BytesWritten ..
		if (inode->c_lnk_len < DidWrite)
		    *BytesWritten = 0;
		else
		    *BytesWritten = WriteSize - (inode->c_lnk_len - DidWrite);
	    }
	}

	CDEBUG ((D_WRAPPER, "coda_write: symlink exit, bytes %d\n",
		 WriteSize));
	return STATUS_SUCCESS;
    }

    if ( coda_get_fileobject(fh, C_O_WRITE) ) {
	CDEBUG((D_WRAPPER, "coda_write: coda_get_fileobject error.\n"));
	return STATUS_INVALID_PARAMETER_1;
    }
    

    //
    // Figure out where to send the IRP
    //
    fsdDevice = IoGetRelatedDeviceObject(inode->c_object);
    
    //
    // Set up our event for use.
    //
    KeInitializeEvent(&event, NotificationEvent, FALSE);
    
    //
    // Allocate and build the IRP we'll be sending to the FSD.
    //    
    irp = IoAllocateIrp(fsdDevice->StackSize, FALSE);    
    if (!irp) {        
	//
	// Allocation failed, presumably due to memory allocation
	// failure.
	//        
	return STATUS_INSUFFICIENT_RESOURCES;        
    }
    
    irp->MdlAddress = Buffer;    
    irp->UserEvent = &event;    
    irp->UserIosb = &iosb;    
    irp->Tail.Overlay.Thread = PsGetCurrentThread();    
    irp->Tail.Overlay.OriginalFileObject= inode->c_object;    
    irp->RequestorMode = KernelMode;
    
    //
    // Indicate that this is a WRITE operation.
    //
    irp->Flags = IRP_WRITE_OPERATION;    
    
    //
    // Set up the next I/O stack location.  These are the parameters
    // that will be passed to the underlying driver.
    //
    ioStackLocation = IoGetNextIrpStackLocation(irp);    
    ioStackLocation->MajorFunction = IRP_MJ_WRITE;    
    ioStackLocation->MinorFunction = 0;    
    ioStackLocation->DeviceObject = fsdDevice;    
    ioStackLocation->FileObject = inode->c_object;
    ioStackLocation->Parameters.Write.Length = WriteSize;    
    ioStackLocation->Parameters.Write.ByteOffset = FileOffset;
    
    //
    // Use the generic completion routine.  This ensures two things:
    //  (1) the MDL we pass won't get freed;
    //  (2) we get I/O completion without the cost.
    //
    IoSetCompletionRoutine(irp, NullDiscardIrpCompletion, 0, TRUE, TRUE,
			   TRUE);
    
    //
    //  I/O operation is built.  Pass it to the underlying file system.
    //
    (void) IoCallDriver(fsdDevice, irp);
    
    //
    // Wait for the I/O to complete.
    //
    KeWaitForSingleObject(&event, Executive, KernelMode, TRUE, 0);
    
    //
    // Copy results from io status block.
    //
    *BytesWritten = iosb.Information;

    //
    // Update the attributes
    //
    if ((FileOffset.QuadPart + *BytesWritten)
	> inode->c_fattr.LastByteWritten.QuadPart)
	inode->c_fattr.LastByteWritten.QuadPart
	    = FileOffset.QuadPart + *BytesWritten;

    //
    // Done.
    //

    CDEBUG ((D_WRAPPER, "coda_write: exit status %d (bytes %d)\n",
	     iosb.Status, *BytesWritten));
    return iosb.Status;
}


static int coda_venusdir_to_fsdir ( IN struct venus_dirent *de,
				    OUT PFS_DIRECTORY_ENTRY Entry,
				    IN char *altname)
{
    char *name;
    NTSTATUS status;
    PUCHAR Ptr;
    ULONG Length;
    ANSI_STRING AnsiString;
    UNICODE_STRING UnicodeString;
    
    
    //
    // pre-set the entry size to zero, in case we need to back-out
    // of this entry
    //
    Entry->EntrySize = 0x00;
    
    //
    // Store the remote File ID where we can later get it
    //
    Entry->Unused.LowPart = 0x0;
    
    //
    // We also use the High Part of the Unused entry as an EOF flag
    //
    Entry->Unused.HighPart = 0x00;
    
    //
    // Get the name.  Since NT wants Wide Characters, we do
    // the conversion here.
    //
    if (altname != NULL)
	name = altname;
    else
	name = de->d_name;  // this is 0 terminated, see coda-src/dir/codadir.c
    
    RtlInitAnsiString(&AnsiString, name);
    status = RtlAnsiStringToUnicodeString(&UnicodeString, &AnsiString,
					  TRUE);
    if ( ! NT_SUCCESS(status) ) {
	CDEBUG((D_FILE, "coda_venusdir_to_fsdir: Cannot convert to "
	       "Unicode!\n"));
	return STATUS_INSUFFICIENT_RESOURCES;
    }
    
    memcpy(&Entry->Name[0], UnicodeString.Buffer, UnicodeString.Length);
    
    //
    // Store the size of the entry
    //
    Entry->NameSize = UnicodeString.Length;
    
    RtlFreeUnicodeString(&UnicodeString);
    
    //
    // Return the NFS "cookie" entry as the DirectoryEntryIndex
    //
    Entry->DirectoryEntryIndex = 0;
    
    //
    // And return the size of the entire entry
    //
    Entry->EntrySize = sizeof(FS_DIRECTORY_ENTRY) + Entry->NameSize;
    
    return 0;   
}


// this converts a Coda style directory buffer to one for the WRAPPER.
// - we return the number of entries found if we are successful
// - we return a negative error if we can't get the attributes


int coda_readdir_rewrite ( coda_inode *dir,
			   unsigned long *DirectoryOffset,
			   void *in,
			   int insize,
			   void *out,
			   int outsize,
			   int count )
{
    NTSTATUS status;
    struct venus_dirent *de;
    char *inp, *end;
    int i;
    char *name;
    int stringoffset = (int) &(((struct venus_dirent *)(0))->d_name);
    UNICODE_STRING UnicodeName;
    ANSI_STRING AnsiName;
    coda_inode *file;
    unsigned long type;
    PFS_FILE_ATTRIBUTES fattr;
    PFS_DIRECTORY_ENTRY Entry = out;
    PFS_DIRECTORY_ENTRY PrevEntry = Entry;
    int Uid = GetNtUID();
    char linkname[CODA_MAXNAMLEN];
    
    // go through the directory and convert existing format into fsd format
    inp = (char *)in;
    end = inp + insize;
    
    i=0;
    while (i < count) {
	
	// some entries go across the buffer boundary,
	// detect those first
	if ((char *)inp + stringoffset > (char *)end) {
	    de = NULL;
	    break;
	}
	
	de = (struct venus_dirent *) inp;
	if ((char *)inp + stringoffset + de->d_namlen + 1
	    > (char *)end ) {
	    de = NULL;
	    break;
	}
	
	// we have the entry, 
	inp += de->d_reclen;
	*DirectoryOffset += de->d_reclen;
	
	// but maybe its the closing entry at the end? 
	if ( de->d_namlen == 0 || inp > end ) {
	    *DirectoryOffset = 0;
	    break;
	}
	
	CDEBUG((D_LOOKATTR, "coda_readdir_rewrite: Entry %d, name %s, "
	       "reclen %d, de 0x%x, end 0x%x\n",
	       i, de->d_name, de->d_reclen, de, end));
	
	// . and .. are not to be reported, skip to the next one 
	if ( ((de->d_namlen ==1)
	      && (RtlCompareMemory(de->d_name, ".", 1) == 1))
	     || ((de->d_namlen ==2)
		 && (RtlCompareMemory(de->d_name, "..",2) == 2))
	     ) {
	    continue;
	}
	
	// Find the file and get the attributes
	status = coda_internal_lookup(dir, de->d_name, de->d_namlen,
				      FALSE, Uid, TRUE, TRUE, &file, NULL); 

	if ( !NT_SUCCESS(status) ) {
	    CDEBUG((D_LOOKATTR, "coda_readdir_rewrite: "
		   "coda_internal_lookup error %s\n", de->d_name));
	    return -STATUS_NO_SUCH_FILE;
	}

	// the entry is fully available, copy it into the output buffer 
	PrevEntry->LastDirectoryEntry = 1;
	if (CODA_ISSYMLINK(file) && (file->c_lnk_state == SLNK_NORMAL)) {
	    strncpy(linkname, de->d_name, CODA_MAXNAMLEN);
	    strncat(linkname, ".lnk", CODA_MAXNAMLEN-de->d_namlen);
	    coda_venusdir_to_fsdir(de, Entry, linkname);
	    CDEBUG((D_LOOKATTR, "coda_readdir_rewrite: "
		   "link name '%s' used\n", linkname));
	} else
	coda_venusdir_to_fsdir(de, Entry, NULL);
	
	fattr = &file->c_fattr;

	Entry->CreationTime = fattr->CreationTime;
	Entry->LastAccessTime = fattr->LastAccessTime;
	Entry->LastModifiedTime = fattr->LastModifiedTime;

#ifdef _BIG_FILE_
       if (file->c_big_filep) {
           CDEBUG((D_WRAPPER, "coda_readdir_rewrite: using big file attrs\n"));
           Entry->ValidDataLength =
               Entry->DiskLength = file->c_big_filep->bf_size;
           Entry->Attributes = FILE_ATTRIBUTE_READONLY;
       } else {
#endif
           Entry->ValidDataLength = fattr->ValidDataLength;
           Entry->DiskLength = fattr->DiskLength;
           Entry->Attributes = fattr->Attributes;
#ifdef _BIG_FILE_
       }
#endif
	Entry->DirectoryEntryIndex = (ULONG) &de;
	
	PrevEntry = Entry;
	Entry = (PFS_DIRECTORY_ENTRY)
	    ((char *)Entry + Entry->EntrySize);
	i++;
    }
    
    EXIT(D_WRAPPER, "coda_readdir_rewrite");
    return i;
}


//
// coda_readdir
//
// - we set *DirectoryOffset to 0 when we encounter the last entry
//     (the last entry is recognized by de + de->rec_len > end 
//      or easier de->d_namlen == 0, see dir/codadir.c, DIR_Convert)  

NTSTATUS coda_readdir(IN FS_FILE_HANDLE DirectoryHandle,
                      IN OUT PULONG DirectoryOffset,
                      IN PFS_DIRECTORY_ENTRY OutputBuffer,
                      IN ULONG OutputBufferSize)
{
    coda_filehandle *dfh = (coda_filehandle *) DirectoryHandle;
    coda_inode *dir = dfh->fh_inode;

    NTSTATUS status;
    ULONG BytesRead;
    ULONG i;
    ULONG EntriesThatWillFit = 0;
    PFS_DIRECTORY_ENTRY Entry = OutputBuffer;
    ULONG Size;
    char *tmpbuf = NULL;
    PMDL pmdl = NULL;
    
    
    CDEBUG ((D_WRAPPER, "coda_readdir: Starting (fh %d: %d/%d)\n",
	     dfh->fh_unique, *DirectoryOffset, OutputBufferSize));

    LockInode(dir);
    
    try {
	
	//
	// First, do some sanity checking to be sure that the handle that
	// was passed in is really a valid directory handle.  For us, this
	// means that it is a pointer to a FILE and marked with TYPE directory.
	//
	if (!CODA_ISOURS(dir) ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: invalid inode!\n"));
	    return(STATUS_INVALID_PARAMETER_1);
	}
	
	if (!CODA_ISDIR(dir) ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: not a directory!\n"));
	    return(STATUS_INVALID_PARAMETER_1);
	}
	
	__try {
	    RtlZeroMemory(OutputBuffer, OutputBufferSize);
	}
	
	except(EXCEPTION_EXECUTE_HANDLER) {
	    
	    CDEBUG((D_WRAPPER, "coda_readdir: Coda readdir can't "
		   "clean buffer!\n"));
	    
	    return(STATUS_INVALID_USER_BUFFER);
	}
	
	
	//
	// Make a worst case determination of the number of entries 
	// that we can read at one time.
	//
	EntriesThatWillFit = OutputBufferSize
	    / (sizeof(FS_DIRECTORY_ENTRY)
	       + CODA_MAXNAMLEN * sizeof(WCHAR));
	
	CDEBUG((D_WRAPPER, "coda_readdir: %d Entries will fit.\n",
	       EntriesThatWillFit));
	
	tmpbuf = coda_malloc(OutputBufferSize);
	if ( !tmpbuf ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: Cannot allocate tmpbuf!\n"));
	    return STATUS_INSUFFICIENT_RESOURCES;
	}
	
	pmdl = IoAllocateMdl(tmpbuf, OutputBufferSize, FALSE, FALSE, NULL);
	if ( !pmdl ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: Cannot allocate MDL\n"));
	    return STATUS_INSUFFICIENT_RESOURCES;
	}
	MmBuildMdlForNonPagedPool(pmdl);
	
	status = coda_read(dfh,
			   RtlConvertUlongToLargeInteger(*DirectoryOffset), 
			   pmdl, OutputBufferSize, &BytesRead);
	
	if(!NT_SUCCESS(status))	{
	    if (status == STATUS_END_OF_FILE) {
		CDEBUG((D_WRAPPER, "coda_readdir: end of directory\n"));
		return STATUS_NO_MORE_FILES;
	    }
	    CDEBUG((D_WRAPPER,
		    "coda_readdir: coda_read failed with %x\n", status));
	    return status;
	}
	
	CDEBUG((D_WRAPPER, "coda_readdir: Read %d bytes from container "
	       "file directory\n", BytesRead));

	status = coda_readdir_rewrite(dir, DirectoryOffset, tmpbuf,
				      BytesRead, OutputBuffer,
				      OutputBufferSize,
				      EntriesThatWillFit);
	
	//  if we didn't read any entries tell the wrapper
	if ( status == 0) {
	    CDEBUG((D_WRAPPER, "coda_readdir: end of directory\n"));
	    return STATUS_NO_MORE_FILES;
	}
	
	//  return bad errors
	if ( status < 0 ) {
	    CDEBUG((D_WRAPPER, "coda_readdir: returning error %d\n", -status));
	    return -status;
	}
	
	CDEBUG((D_WRAPPER, "coda_readdir: Returning Offset = %d, "
		"entries read = %d\n", *DirectoryOffset, status));
    } 
    
    finally {
	if (pmdl)
	    IoFreeMdl(pmdl);
	if (tmpbuf)
	    coda_free(tmpbuf);
	UnLockInode(dir);
    }
    
    EXIT(D_WRAPPER, "coda_readdir");
    return(STATUS_SUCCESS);
}


//
// Do Nothing - this is used to make sure coda_remshare gets called on close().
//

NTSTATUS coda_updshare (IN FS_FILE_HANDLE FileHandle,
			IN ULONG Access)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;    
    CDEBUG ((D_WRAPPER, "coda_updshare: fh %d Access %x %s\n", fh->fh_unique,
	     Access, (fh->fh_isopen ? "is open" : "")));

    return  STATUS_SUCCESS;
}

// We are using this as a "Close" operation.  It appears to be
// called at every file close.

NTSTATUS coda_remshare (IN FS_FILE_HANDLE FileHandle,
			IN ULONG Access)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode *inode = fh->fh_inode;
    struct coda_ntvolume *ntvol;
    NTSTATUS rc;
    int flags = 0;

    CDEBUG ((D_WRAPPER, "coda_remshare: fh %d Access %x %s\n", fh->fh_unique,
	     Access, (fh->fh_isopen ? "is open" : "")));

    if ( !CODA_ISOURS(inode) ) {
	CDEBUG((D_WRAPPER, "coda_remshare: Bad file %s\n"));
	return STATUS_SUCCESS;
    }
    
    if ( !CODA_ISDIR(inode) && fh->fh_isopen
	 && (fh->fh_openflags & C_O_WRITE)) {
	// Forces writes to happen
	
	rc = OwFlushCache (FileHandle, 0);
	if (rc != STATUS_SUCCESS) {
	    CDEBUG((D_WRAPPER,
		    "coda_remshare: OwFlushCache returns error %d.\n",
		    rc));
	}
    }

    // Now "close" the file.
    
    if (fh->fh_isopen) {
	
	CDEBUG ((D_WRAPPER, "coda_remshare: Closing fh %d\n", fh->fh_unique));
	
	LockInode (inode);
	
	if (!(--(inode->c_open_cnt))) {
	    CDEBUG((D_FILE, "coda_remshare: Closing inode file object.\n")); 
	    // Close the object!
	    if (inode->c_object != NULL) {	
		inode_close_handle(inode);
	    } else {
		DbgPrint("coda_remshare: Closing NULL c_object\n");
	    }
	}
	
	rc = venus_close (&inode->c_fid, fh->fh_openflags, fh->fh_uid);
	
	// ignore error
	if ( rc ) {
	    DbgPrint("coda_remshare: venus_close error %d\n", rc);
	}
	
	// Update open flags
	fh->fh_openflags = 0;
	fh->fh_isopen = 0;
	UnLockInode(inode);
    }

#ifdef _BIG_FILE_
    if (!(inode->c_fstype & FILE_ATTRIBUTE_DIRECTORY)
	&& (strcmp(inode->c_name, "_Coda_BigFile_") == 0)) {

	coda_inode *parentinode = inode->c_parent;

	CDEBUG((D_WRAPPER, "coda_remshare: Closing _Coda_BigFile_\n"));

	// purge cache for parent inode (the possible big file) and its parent.
	purge_fh_list (parentinode);
	purge_fh_list (parentinode->c_parent);

	// Say to throw away the inode so we can start fresh
	inode_unsetflag(parentinode, I_PURGE);
    }
#endif
    
    EXIT(D_WRAPPER, "coda_remshare");
    return STATUS_SUCCESS;
}


//
// coda_create
//
//   This routine is similar to lookup, except that the specified mode
//   for the file is to create it, rather than open it.
//
// Inputs:
//   DirectoryHandle - the directory where this file is being created.
//   FileName - the file to create
//   ShortFileName - unused (short name generated by FSDK.)
//   FileType - file/directory being created
//
// Outputs:
//   NewFileHandle - the newly created file handle.
//
// Returns:
//    SUCCESS - the operation was successful.
//    OBJECT_NAME_EXISTS - the specified name already exists.
//
// Notes:
//   None.
//


NTSTATUS coda_create (IN FS_FILE_HANDLE DirectoryHandle,
		      IN PWSTR FileName,
		      IN PWSTR ShortFileName,
		      IN ULONG FileType,
		      OUT PFS_FILE_HANDLE NewFileHandle)
     
{
    NTSTATUS status = STATUS_SUCCESS;
    UNICODE_STRING unicodeFileName;
    ANSI_STRING AnsiFileName;

    HANDLE fileHandle;
    PFILE_OBJECT fileObject;
    FS_FILE_ATTRIBUTES  tattr;  // Temp variable for getting attributes.
    
    CHAR NormalFileName[CODA_MAXNAMLEN];
    int  namelen;
    coda_filehandle *dfh = (coda_filehandle *) DirectoryHandle;
    coda_filehandle *nfh ;
    coda_inode *dirinode = dfh->fh_inode;
    coda_inode *newinode;
    CodaFid newfid;
    struct coda_vattr attrs;
    int rc;	
    int Uid = GetNtUID ();
    int NewLink = FALSE;
    
    CDEBUG ((D_WRAPPER, "coda_create: dir fh %d\n", dfh->fh_unique));
    
    //
    // First, sanity checking
    //
    
    //
    // Sanity checking
    //
    if ( !CODA_ISOURS(dirinode) ) {
	DbgPrint ("Bad file %s\n");
	return STATUS_INVALID_PARAMETER_1;
    }

    if ( !CODA_ISDIR(dirinode) ) {
	return STATUS_NOT_A_DIRECTORY;
    }

    LockInode (dirinode);

    //
    // Convert from WCHAR to ASCII via a unicode string.
    //

    unicodeFileName.Buffer = FileName;
    unicodeFileName.Length = wcslen (FileName) * sizeof (WCHAR);
    unicodeFileName.MaximumLength = unicodeFileName.Length + sizeof (WCHAR);
    RtlUnicodeStringToAnsiString (&AnsiFileName, &unicodeFileName, TRUE);
    if ( AnsiFileName.Length > CODA_MAXNAMLEN ) {
	UnLockInode (dirinode);
	return STATUS_INVALID_PARAMETER_2;
    }
    memcpy (NormalFileName, AnsiFileName.Buffer, AnsiFileName.Length);
    namelen = AnsiFileName.Length;
    NormalFileName[namelen] = 0x0;
    RtlFreeAnsiString (&AnsiFileName);
    
    //
    // Tell venus to create the file.  Don't open the container until needed.
    //
	
    if ( FileType != OW_HANDLE_TYPE_DIRECTORY ) {
	int mode = 0666;

	CDEBUG ((D_WRAPPER, "coda_create: create file %s\n", NormalFileName));

	// Special cases ... make .exe file executible
	if (strcasetail(NormalFileName, ".exe"))
	    mode = 0777;

	if (strcasetail(NormalFileName, ".lnk")) {
	    
	    CDEBUG ((D_WRAPPER, "coda_create: create .lnk file %s\n",
		     NormalFileName));
	    NewLink = TRUE;
	}
	
	rc = venus_create (&dirinode->c_fid, NormalFileName, namelen, 1,
			   mode, 0, &newfid, &attrs, dfh->fh_uid);
	
    } else {
	CDEBUG ((D_WRAPPER, "coda_create: create dir %s\n", NormalFileName));

	rc = venus_mkdir (&dirinode->c_fid, NormalFileName,
			  namelen, &newfid, &attrs, dfh->fh_uid);
    }

    if ( rc ) {
	UnLockInode (dirinode);
	return STATUS_ACCESS_DENIED;
    }
    
    // Now instantiate the new inode

    GrabLock ();

    CDEBUG ((D_WRAPPER, "coda_create: new fid =%s\n", coda_f2s(&newfid)));

    newinode =  inode_new (&newfid, dirinode, NormalFileName);
    if ( !newinode ) {
	CDEBUG ((D_WRAPPER, "coda_create: inode_new failed!!!\n"));
	UnLockInode (dirinode);
	DropLock ();
	return STATUS_INVALID_PARAMETER_1;
    }

    UnLockInode (dirinode);

    // Create the fh for the new file/directory

    nfh = get_cfh (newinode, dirinode, Uid);

    DropLock ();

    if ( !nfh ) {
	CDEBUG ((D_CNODE, "coda_create: get_cfh failed!!!\n"));
	inode_release (newinode);
	return STATUS_INVALID_PARAMETER_1;
    }

    if (NewLink) {
	newinode->c_lnk_len = 0;
	newinode->c_lnk_state = SLNK_NEW;
    }

    // Get the attributes
    rc = coda_internal_getattr (newinode, nfh->fh_uid, FALSE,
				&newinode->c_fattr);
    if ( rc )
	DbgPrint ("coda_create: could not get attrs on new file/dir!\n");

    *NewFileHandle = (PFS_FILE_HANDLE)nfh;
  
    CDEBUG ((D_WRAPPER, "coda_create: returning fh %d\n", nfh->fh_unique));
    return STATUS_SUCCESS;
}



//
//
// coda_delete2 -- New version of delete
//

NTSTATUS coda_delete2(IN FS_FILE_HANDLE FileHandle,
		      IN FS_FILE_HANDLE ParentHandle,
		      IN PWSTR FileName)
{
    static delaydelete = 1;
    NTSTATUS Status;
    coda_filehandle *dfh = (coda_filehandle *) ParentHandle;
    coda_inode *dirinode = dfh->fh_inode;
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode *finode = fh->fh_inode;

    int rc;
    
    CDEBUG ((D_WRAPPER, "coda_delete2: dir fh %d, fh %d\n",
	    dfh->fh_unique, fh->fh_unique));

    LockInode(dirinode);
    
    try {

	if (!CODA_ISOURS(finode)) {
	    return STATUS_INVALID_PARAMETER_1;
	}

	if (!CODA_ISOURS(dirinode)) {
	    return STATUS_INVALID_PARAMETER_2;
	}
	
	if (!CODA_ISDIR(dirinode)) {
	    return STATUS_NOT_A_DIRECTORY;
	}
	
	if (dirinode != finode->c_parent) {
	    return STATUS_CANNOT_DELETE;
	}

	// Inode contains name use it.  Wrapper can possibly
	// send NULL for FileName, so don't use it.

#ifdef _BIG_FILE_
	if (finode->c_big_filep) {
	    coda_filehandle *bffh;
	    coda_inode *bfinode;
	    NTSTATUS rc;

	    // Find the magic file
	    rc = coda_internal_lookup (finode, "_Coda_BigFile_", 14, FALSE,
				       fh->fh_uid, FALSE, FALSE,
				       &bfinode, &bffh);
	    if (!NT_SUCCESS(rc)) {
		DbgPrint ("coda_delete2: Can't find _Coda_BigFile_!\n");
		return STATUS_CANNOT_DELETE;
	    }

	    // Delete the magic file ... we don't use the file name ...
	    rc = coda_delete2 (bffh, fh, NULL);

	    // If successful, release the parent's file handle also
	    if (NT_SUCCESS(rc)) {
		// Turn off the big file attribute
		inode_setflag(finode, I_PURGE);
		coda_release(FileHandle);
	    } else {
		// coda_delete2 didn't release bffh, do it here
		inode_setflag(bffh->fh_inode, I_PURGE);
		coda_release((FS_FILE_HANDLE)bffh);
	    }

	    return rc;
	}
#endif

	LockInode(finode);
	
	if (!CODA_ISDIR(finode) && finode->c_fh_cnt > 1) {
	    // multiple file handles ... do an open to force
	    // the file to stay available .... 
	    // Just delete the directories!
	    char newname [CODA_MAXNAMLEN];
	    CDEBUG((D_WRAPPER, "coda_delete2: opening file...\n"));

	    rc = venus_open_by_path (&(finode->c_fid), C_O_WRITE, newname,
				     fh->fh_uid);

	    if (!rc) {
		inode_setflag(finode, I_DELETE);
		inode_cache_delete (finode);
	    } else {
		UnLockInode(finode);
		return STATUS_CANNOT_DELETE;
	    }
	}
	if (CODA_ISDIR(finode)) {
	    rc = venus_rmdir (&(dirinode->c_fid), finode->c_name,
			      strlen(finode->c_name), fh->fh_uid);
	} else {
	    rc = venus_remove (&(dirinode->c_fid), finode->c_name,
			       strlen(finode->c_name), fh->fh_uid);
	}

	if (rc)
	    Status = STATUS_CANNOT_DELETE;
	else
	    Status = STATUS_SUCCESS;

	UnLockInode(finode);

        if(NT_SUCCESS(Status) )
            {
		CDEBUG ((D_WRAPPER, "coda_delete2: calling coda_release\n"));
		inode_setflag(finode, I_PURGE);
		Status = coda_release(FileHandle);
            }
    } 

    finally {
        UnLockInode(dirinode);
    }
    
    EXIT (D_WRAPPER, "coda_delete2");
    return(Status);
}

//
// coda_rename2 ... rename a file
//

NTSTATUS coda_rename2(IN FS_FILE_HANDLE SourceDir,
		      IN FS_FILE_HANDLE TargetDir,
		      IN FS_FILE_HANDLE SourceHandle,
		      IN PWSTR SourceName,
		      IN PWSTR TargetName,
		      IN PWSTR AltTargetName,
		      OUT PFS_FILE_HANDLE TargetHandle)
{
    coda_filehandle *srcdfh = (coda_filehandle *) SourceDir;
    coda_inode *srcdir = srcdfh->fh_inode;
    coda_filehandle *trgdfh = (coda_filehandle *) TargetDir;
    coda_inode *trgdir = trgdfh->fh_inode;
    coda_filehandle *srcfh = (coda_filehandle *) SourceHandle;
    coda_inode *srcfile = srcfh->fh_inode;
    //    coda_filehandle *trgfh;
    //    coda_inode *trgfile;

    char srcName [CODA_MAXNAMLEN];
    char trgName [CODA_MAXNAMLEN];

    UNICODE_STRING unicodeFileName;
    ANSI_STRING AnsiFileName;

    int rc;

    CDEBUG ((D_WRAPPER, "coda_rename: src dfh %d, trg dfh %d, fh %d\n",
	     srcdfh->fh_unique, trgdfh->fh_unique, srcfh->fh_unique));
   
    if (!CODA_ISOURS(srcdir)) {
	return STATUS_INVALID_PARAMETER_1;
    }
    
    if (!CODA_ISOURS(trgdir)) {
	return STATUS_INVALID_PARAMETER_2;
    }

    if (!CODA_ISOURS(srcfile) || srcfile->c_parent != srcdir) {
	return STATUS_INVALID_PARAMETER_3;
    }

    // Lock Things!
    LockInode (srcdir);
    LockInode (trgdir);
    LockInode (srcfile);

    try {
	// Make the strings.....
	unicodeFileName.Buffer = SourceName;
	unicodeFileName.Length = wcslen(SourceName) * sizeof(WCHAR);
	unicodeFileName.MaximumLength = unicodeFileName.Length + sizeof(WCHAR);
	RtlUnicodeStringToAnsiString(&AnsiFileName, &unicodeFileName, TRUE);
	if(AnsiFileName.Length > CODA_MAXNAMLEN) {
	    return(STATUS_INVALID_PARAMETER_2);
	}
	memcpy(srcName, AnsiFileName.Buffer, AnsiFileName.Length);
	srcName[AnsiFileName.Length] = 0x0;
	RtlFreeAnsiString(&AnsiFileName);
	
	unicodeFileName.Buffer = TargetName;
	unicodeFileName.Length = wcslen(TargetName) * sizeof(WCHAR);
	unicodeFileName.MaximumLength = unicodeFileName.Length + sizeof(WCHAR);
	RtlUnicodeStringToAnsiString(&AnsiFileName, &unicodeFileName, TRUE);
	if(AnsiFileName.Length > CODA_MAXNAMLEN) {
	    return(STATUS_INVALID_PARAMETER_2);
	}
	memcpy(trgName, AnsiFileName.Buffer, AnsiFileName.Length);
	trgName[AnsiFileName.Length] = 0x0;
	RtlFreeAnsiString(&AnsiFileName);
	
	rc = venus_rename (&(srcdir->c_fid), srcName, strlen(srcName),
			   &(trgdir->c_fid), trgName, strlen(trgName),
			   srcdfh->fh_uid);
	
	if (rc) {
	    return STATUS_INVALID_HANDLE;
	}

	// inode manipulations ....
	inode_rename (srcfile, trgdir, trgName);
	
	*TargetHandle = (FS_FILE_HANDLE) srcfh;
    }

    finally {
	UnLockInode (srcdir);
	UnLockInode (trgdir);
	UnLockInode (srcfile);
    }

    EXIT (D_WRAPPER, "coda_rename");

    return STATUS_SUCCESS;
}


//
// coda_access -- this appears where the coda security must be handled!
//

NTSTATUS coda_access(IN FS_FILE_HANDLE FileHandle,
		     IN OUT PACCESS_STATE AccessState)
{
    coda_filehandle *fh = (coda_filehandle *) FileHandle;
    coda_inode *inode = fh->fh_inode;
    int flags;
    int i;
    ULONG mask;
    int rc;

    CDEBUG ((D_WRAPPER, "coda_access: name = '%s' %s fh %d, Prev = 0x%x, "
	     "Remain = 0x%x fh_granted = 0x%x\n", inode->c_name,
	     coda_f2s(&inode->c_fid), fh->fh_unique,
	     AccessState->PreviouslyGrantedAccess,
	     AccessState->RemainingDesiredAccess, fh->fh_granted));

    // Checking for now ... Should we do something?
    if (fh->fh_isopen) {
	CDEBUG ((D_WRAPPER, "coda_access: FH is open.  No action done.\n"));
    }

    // check for root directory and set no caching
    if (!root_init && fh == root_fh) {
	if (OwDisableFileCache(FileHandle,TRUE) != STATUS_SUCCESS)
	    CDEBUG((D_WRAPPER,
		    "coda_access: could not flush root dir cache\n"));
	root_init = TRUE;
    }

    // check access
    if ((fh->fh_granted & AccessState->RemainingDesiredAccess) ==
	AccessState->RemainingDesiredAccess) {

	CDEBUG ((D_WRAPPER, "coda_access: already has that access.\n"));

	AccessState->PreviouslyGrantedAccess |= 
	    AccessState->RemainingDesiredAccess;
	AccessState->RemainingDesiredAccess = 0;
	fh->fh_lastacc = AccessState->PreviouslyGrantedAccess;
        return STATUS_SUCCESS;	
    }

    LockInode (inode);

    if (AccessState->RemainingDesiredAccess) {
	
	// Create a coda flags with the proper requests.
	flags = 0;
	
	if (AccessState->RemainingDesiredAccess
	    & (FILE_READ_DATA | FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES
	       | GENERIC_READ))
	    flags |= C_A_R_OK;
	if (AccessState->RemainingDesiredAccess
	    & (FILE_WRITE_DATA | FILE_APPEND_DATA | FILE_WRITE_ATTRIBUTES
	       | FILE_DELETE_CHILD | GENERIC_WRITE ))
	    flags |= C_A_W_OK;
	if (AccessState->RemainingDesiredAccess &
	    (FILE_EXECUTE | GENERIC_EXECUTE))
	    flags |= C_A_X_OK; 
	if (AccessState->RemainingDesiredAccess
	    & (FILE_ADD_FILE | FILE_ADD_SUBDIRECTORY))
	    flags |= C_A_C_OK;

#ifdef _BIG_FILE_
	// Check only read access for big files.
	if (inode->c_big_filep && (flags & (C_A_C_OK | C_A_X_OK | C_A_W_OK))) {
	    CDEBUG ((D_WRAPPER,
		     "coda_access: returning access denied (big file)\n"));
	    UnLockInode (inode);
	    return STATUS_ACCESS_DENIED;
	}
#endif
	if ( CODA_ISSYMLINK(inode) && inode->c_containername[0] == '\1' &&
	     ((flags & (C_A_R_OK | C_A_W_OK)) == (C_A_R_OK | C_A_W_OK)) ) {
	    CDEBUG ((D_UPCALL, "coda_access: granting symlink write\n"));
	} else
	    if (fh != coda_nt_volume.v_ctlfh) {
		
		CDEBUG ((D_UPCALL, "venus_access: flags=%x\n", flags));
		
		rc = venus_access (&(inode->c_fid), flags, fh->fh_uid);
		
		if (rc) {
		    CDEBUG ((D_WRAPPER,
			     "coda_access: returning access denied\n"));
		    UnLockInode (inode);
		    return STATUS_ACCESS_DENIED;
		}
	    }
	
	mask = AccessState->RemainingDesiredAccess 
	    & (FILE_READ_DATA | FILE_LIST_DIRECTORY | FILE_READ_ATTRIBUTES
	       | FILE_GENERIC_READ | FILE_WRITE_DATA | FILE_APPEND_DATA
	       | FILE_WRITE_ATTRIBUTES | FILE_DELETE_CHILD | FILE_GENERIC_WRITE
	       | FILE_EXECUTE | FILE_GENERIC_EXECUTE | FILE_ADD_FILE
	       | FILE_ADD_SUBDIRECTORY);
	
	AccessState->PreviouslyGrantedAccess |=
	    AccessState->RemainingDesiredAccess & mask;
	
	AccessState->RemainingDesiredAccess &= ~mask;
    }

    // Only add to the granted access rights.
    fh->fh_granted |= AccessState->PreviouslyGrantedAccess;
    fh->fh_lastacc = fh->fh_granted;

    UnLockInode (inode);

    CDEBUG ((D_WRAPPER, "coda_access: returning access granted 0x%x\n",
	     fh->fh_granted));

    return STATUS_SUCCESS;
}


// Normally called only on the CTL handle.

NTSTATUS coda_ioctl(IN FS_FILE_HANDLE FileHandle, IN ULONG ControlCode,
		    IN PVOID InputBuffer, IN ULONG InputBufferSize,
		    IN PVOID OutputBuffer, IN ULONG OutputBufferSize,
		    OUT PULONG BytesWritten, OUT PULONG BytesNeeded)
{

    NTSTATUS Status = STATUS_SUCCESS;

    CDEBUG((D_WRAPPER, "coda_ioctl: Starting, code: %x\n", ControlCode));
    
    switch (ControlCode) {
	
    case CODA_FSCTL_PIOCTL: 
	/// 
	/// Pioctl interface.  
	///
	{
	    ULONG cmd;
	    void *buf;
	    ULONG bufsize;
	    ULONG outsize = OutputBufferSize;

	    // Verify it is the correct fh ...
	    if (FileHandle != coda_nt_volume.v_ctlfh) {
		CDEBUG ((D_WRAPPER, "coda_ioctl: pioctl on wrong fh\n"));
		Status = STATUS_INVALID_DEVICE_REQUEST;
		break;
	    }
 
	    bufsize = MAX(InputBufferSize, OutputBufferSize);
	    
	    Status = STATUS_INVALID_DEVICE_REQUEST;
	    buf = coda_malloc(bufsize);
	    if  ( !buf  ) {
		DbgPrint("No memory for request data.\n");
		break;
	    }
	    
	    RtlCopyMemory(buf, InputBuffer, InputBufferSize);
	    
	    // Do the pioctl!
	    Status = coda_pioctl ((struct PioctlData*) buf, InputBufferSize,
				  OutputBuffer, &outsize);
	    
	    if ( Status ) {
		DbgPrint("coda_fsctrl: coda_pioctl returns error %d\n",
			 Status);
	    }
	    *BytesWritten = outsize;
	    coda_free (buf);
	}
	break;
	
    default:
	DbgPrint("coda_fsctrl: Invalid request %d\n", ControlCode);
	Status = STATUS_INVALID_DEVICE_REQUEST;
    }
    
    EXIT(D_WRAPPER, "coda_ioctl");
    return Status;
}


//
// Pseudo wrapper call call ... actually called from a CodaFsCtrl ...
//   but in UNIX this is a VFS call .... and thus like a wrapper call!
//

NTSTATUS coda_pioctl (IN struct PioctlData *buf, IN ULONG bufsize,
		      IN PVOID outbuf, IN OUT PULONG poutsize)
{
    char *FilePath = ((char *)(buf)) + sizeof (struct PioctlData);
    int   PathLen = strlen(FilePath);
    CodaFid *pfid;
    coda_inode *inode;
    int rc;
    int Uid = GetNtUID();

    CDEBUG( (D_PIOCTL, "coda_pioctl: Information!\n"));
    CDEBUG( (D_PIOCTL, "    FilePath = '%s', PathLen = %d\n", FilePath,
	    PathLen));
    CDEBUG( (D_PIOCTL, "    cmd = 0x%x\n", buf->cmd));
    CDEBUG( (D_PIOCTL, "    in_size = %d  out_size = %d \n", buf->vi.in_size,
	    buf->vi.out_size));

    // Fill in the buf->vi.in with the correct pointer!
    buf->vi.in = (caddr_t) (FilePath + PathLen + 1);

    // Find the file by name ....  Need to turn \ -> / ???  May not.
    if (PathLen) {

	// Look it up!  XXX May need more than this! 
	rc = coda_internal_path_lookup (coda_nt_volume.v_rootinode, FilePath,
					FALSE, Uid, &inode);
	if (rc != STATUS_SUCCESS) {
	    CDEBUG( (D_PIOCTL,
		    "coda_pioctl: coda_internal_path_lookup error\n"));
	    *poutsize = 0;
	    return STATUS_INVALID_DEVICE_REQUEST;
	}
    } else {
	// Use the root fid!
	inode = coda_nt_volume.v_rootinode;
    }

    // Ask venus

    rc = venus_pioctl (&(inode->c_fid), buf, bufsize, outbuf, poutsize, Uid);

    CDEBUG( (D_PIOCTL, "coda_pioctl: returning, venus_pioctl rc = %d\n", rc));

    if (rc)
	return STATUS_INVALID_DEVICE_REQUEST;

    return STATUS_SUCCESS;
}
